
import os
import ast
import time
from enum import Enum
from typeguard import typechecked

from benchmark_runner.common.logger.logger_time_stamp import logger_time_stamp, logger
from benchmark_runner.common.oc.oc_exceptions import (PodNotCreateTimeout, PodNotInitializedTimeout, PodNotReadyTimeout, \
    PodNotCompletedTimeout, PodTerminateTimeout, PodNameNotExist, LoginFailed, VMNotCreateTimeout, VMDeleteTimeout, \
    YAMLNotExist, VMNameNotExist, VMNotInitializedTimeout, VMNotReadyTimeout, VMStateTimeout, VMNotCompletedTimeout, \
    ExecFailed, PodFailed, DVStatusTimeout, CSVNotCreateTimeout, UpgradeNotStartTimeout, OperatorInstallationTimeout, OperatorUpgradeTimeout)
from benchmark_runner.common.ssh.ssh import SSH
from benchmark_runner.main.environment_variables import environment_variables


class VMStatus(Enum):
    Stopped = 'Stopped'
    Starting = 'Starting'
    Running = 'Running'


class OC(SSH):
    """
    This class run OC commands
    """

    SHORT_TIMEOUT = 600
    # sleep time between checks is 5 sec
    SLEEP_TIME = 3

    def __init__(self, kubeadmin_password: str = ''):
        super().__init__()
        self.__kubeadmin_password = kubeadmin_password
        self.__environment_variables_dict = environment_variables.environment_variables_dict
        self._run_artifacts = self.__environment_variables_dict.get('run_artifacts_path', '')
        self.__elasticsearch_url = self.__environment_variables_dict.get('elasticsearch_url', '')
        self.__kata_csv = self.__environment_variables_dict.get('kata_csv', '')
        self.__cli = self.__environment_variables_dict.get('cli', '')
        self.__worker_disk_prefix = self.__environment_variables_dict.get('worker_disk_prefix', '')
        self.__worker_disk_ids = self.__environment_variables_dict.get('worker_disk_ids', '')
        if self.__worker_disk_ids:
            self.__worker_disk_ids = ast.literal_eval(self.__worker_disk_ids)

    def get_ocp_server_version(self):
        """
        This method returns ocp server version
        :return:
        """
        return self.run(f"{self.__cli} get clusterversion version -o jsonpath='{{.status.desired.version}}'")

    def upgrade_ocp(self, upgrade_ocp_version: str):
        """
        This method upgrades OCP version with conditional handling for specific versions.

        @param upgrade_ocp_version: Version to upgrade to
        @return:
        """
        ocp_channel = '.'.join(upgrade_ocp_version.split('.')[:2])

        # see: https://access.redhat.com/articles/7031404
        if ocp_channel == "4.16":
            patch_command = f"{self.__cli} -n openshift-config patch cm admin-acks --patch '{{\"data\":{{\"ack-4.15-kube-1.29-api-removals-in-4.16\":\"true\"}}}}' --type=merge"
            self.run(patch_command)

        upgrade_command = f"{self.__cli} adm upgrade ; sleep 10; {self.__cli} adm upgrade channel stable-{ocp_channel}; sleep 10; {self.__cli} adm upgrade --to={upgrade_ocp_version};"
        self.run(upgrade_command)

    def upgrade_in_progress(self):
        """
        This method returns True when an upgrade is in progress and False when it is not.
        @return: bool
        """
        status = self.run(f"{self.__cli} get clusterversion version -o jsonpath='{{.status.conditions[?(@.type==\"Progressing\")].status}}'")
        return status == 'True'

    @logger_time_stamp
    def wait_for_upgrade_start(self, upgrade_version: str, timeout: int = SHORT_TIMEOUT):
        """
        This method waits for ocp upgrade to start
        :param upgrade_version:
        :param timeout:
        :return:
        """
        current_wait_time = 0
        while timeout <= 0 or current_wait_time <= timeout and not self.upgrade_in_progress():
            # sleep for x seconds
            time.sleep(OC.SLEEP_TIME)
            current_wait_time += OC.SLEEP_TIME
        if self.upgrade_in_progress():
            return True
        else:
            raise UpgradeNotStartTimeout(version=upgrade_version)

    def get_upgrade_version(self):
        """
        This method returns upgrade version
        @return:
        """
        return self.run(f"{self.__cli} get clusterversion version -o jsonpath='{{.status.desired.version}}'")

    def get_cluster_status(self):
        """
        This method returns the STATUS from the 'oc get clusterversion' command.
        @return: str - The current status of the cluster version.
        """
        return self.run(f"{self.__cli} get clusterversion version -o jsonpath='{{.status.conditions[?(@.type==\"Progressing\")].message}}'")

    def get_operator_version(self, namespace):
        """
        This method returns the operator version from the specified namespace.
        @param namespace: str - The namespace to search for the operator version.
        @return: major version
        """
        version = self.run(f"{self.__cli} get csv -n {namespace} -o jsonpath='{{.items[0].spec.version}}'")
        return '.'.join(version.split('.')[:2])

    def wait_for_operator_installation(self, operator: str, version: str, namespace: str, timeout: int = SHORT_TIMEOUT):
        """
        This method waits till operator version is installed successfully
        @param operator:
        @param version:
        @param timeout:
        @param namespace:
        @return:
        """
        current_wait_time = 0
        while timeout <= 0 or current_wait_time <= timeout and not self.get_operator_version(namespace) == version:
            # sleep for x seconds
            time.sleep(OC.SLEEP_TIME)
            current_wait_time += OC.SLEEP_TIME
        if self.get_operator_version(namespace) == version:
            logger.info(f'{operator} operator version: {version} in namespace: {namespace} has been installed successfully')
            return True
        else:
            raise OperatorInstallationTimeout(operator=operator, version=version, namespace=namespace)

    def healthcheck(self, action: str):
        """
        This method stops/resumes ocp health check according to action
        @param action:
        @return:
        """
        if action == 'stop':
            self.run(f"{self.__cli} -n openshift-machine-api annotate mhc $({self.__cli} get machinehealthcheck -n openshift-machine-api -o jsonpath='{{.items[0].metadata.name}}') cluster.x-k8s.io/paused=\"\"")
        elif action == 'resume':
            self.run(f"{self.__cli} -n openshift-machine-api annotate mhc $({self.__cli} get machinehealthcheck -n openshift-machine-api -o jsonpath='{{.items[0].metadata.name}}') cluster.x-k8s.io/paused-")

    def get_cnv_version(self):
        """
        This method returns cnv version
        :return:
        """
        return self.run(f"{self.__cli} get csv -n openshift-cnv $({self.__cli} get csv -n openshift-cnv --no-headers | awk '{{ print $1; }}') -o jsonpath='{{.spec.version}}'")

    def get_odf_version(self):
        """
        This method returns odf version
        :return:
        """
        return self.run(f"{self.__cli} get csv -n openshift-storage -o jsonpath='{{.items[0].spec.labels.full_version}}'")

    def remove_lso_path(self):
        """
        The method removes lso path on each node
        @return:
        """
        self.run(fr"""{self.__cli} get nodes -l node-role.kubernetes.io/worker= -o jsonpath="{{range .items[*]}}{{.metadata.name}}{{'\\n'}}{{end}}" |  xargs -I{{}} {self.__cli} debug node/{{}} -- chroot /host sh -c "rm -rfv /mnt/local-storage/local-sc/" """)

    def get_worker_disk_ids(self, node: str = None):
        """
        This method returns all worker disk ids or disk ids by node
        :param node:
        :return:
        """
        workers_disk_ids = []

        if node and node in self.__worker_disk_ids:
            workers_disk_ids.extend(self.__worker_disk_ids[node])
        else:
            for node, disk_ids in self.__worker_disk_ids.items():
                workers_disk_ids.extend(disk_ids)

        return workers_disk_ids

    def get_pv_disk_ids(self):
        """
        This method returns list of pv disk ids
        """
        pv_ids = self.run(
            f"{self.__cli} get pv -o jsonpath='{{range .items[*]}}{{.metadata.annotations.storage\\.openshift\\.com/device-id}}{{\"\\n\"}}{{end}}'")
        return [pv[len(self.__worker_disk_prefix):] for pv in pv_ids.split()]

    def get_free_disk_id(self, node: str = None):
        """
        This method returns free disk per node  [workers disk ids - workers pv disk ids]
        """
        workers_disk_ids = self.get_worker_disk_ids(node)
        workers_pv_disk_ids = self.get_pv_disk_ids()
        free_disk_id = f'{self.__worker_disk_prefix}{[disk_id for disk_id in workers_disk_ids if disk_id not in workers_pv_disk_ids][0]}'
        if free_disk_id:
            return free_disk_id
        else:
            raise ValueError('Missing free disk id')

    @typechecked
    def run_debug_node(self, node: str, cmd: str):
        """
        This method runs command in debug node
        :param: node
        :param: cmd
        :return:
        """
        self.run(f"{self.__cli} debug node/{node} --no-tty=true -- chroot /host sh -c '{cmd}'")

    def get_kata_operator_version(self):
        """
        This method returns kata operator version
        :return:
        """
        return self.run(f"{self.__cli} get csv -n openshift-sandboxed-containers-operator $({self.__cli} get csv -n openshift-sandboxed-containers-operator --no-headers | awk '{{ print $1; }}') -o jsonpath='{{.spec.version}}'")

    @typechecked
    def get_kata_rpm_version(self, node: str):
        """
        This method returns sandboxed containers (kata) rpm version
        @param: node
        :return:
        """
        kata_rpm_version = self.run(f"{self.__cli} debug node/{node} -- chroot /host rpm -q --queryformat='%{{VERSION}}-%{{RELEASE}}' kata-containers 2>/dev/null")
        return '.'.join(kata_rpm_version.split('.')[:3])

    def _get_kata_default_channel(self):
        """
        This method retrieves the default channel for Kata
        """
        return self.run(f"{self.__cli} get packagemanifest -n openshift-marketplace sandboxed-containers-operator -o jsonpath='{{.status.defaultChannel}}'")

    def _get_kata_default_channel_field(self, channel_field: str):
        """
        This method retrieves a field from the packagemanifest for the default Kata channel
        """
        default_channel = f'"{self._get_kata_default_channel()}"'
        command = f"{self.__cli} get packagemanifest -n openshift-marketplace sandboxed-containers-operator -ojson | jq -r '[foreach .status.channels[] as $channel ([[],[]];0;(if ($channel.name == {default_channel}) then $channel.{channel_field} else null end))] | flatten | map (select (. != null))[]'"
        return self.run(command)

    def _get_kata_csv(self):
        """
        This method retrieves the CSV of the sandboxed containers operator for installation"
        """
        return self._get_kata_default_channel_field("currentCSV")

    def _get_kata_catalog_source(self):
        """
        This method retrieves the catalog source of the sandboxed containers operator for installation"
        """
        return self.run(f"{self.__cli} get packagemanifest -n openshift-marketplace sandboxed-containers-operator -o jsonpath='{{.status.catalogSource}}'")

    def _get_kata_channel(self):
        """
        This method retrieves the channel of the sandboxed containers operator for installation"
        """
        return self._get_kata_default_channel_field("name")

    def _get_kata_namespace(self):
        """
        This method retrieves the namespace of the sandboxed containers operator for installation"
        """
        return self._get_kata_default_channel_field('currentCSVDesc.annotations."operatorframework.io/suggested-namespace"')

    def set_kata_threads_pool(self, thread_pool_size: str):
        """
        This method sets kata thread-pool-size in every worker node
        @param thread_pool_size:
        @return:
        """
        self.run(fr"""{self.__cli} get nodes -l node-role.kubernetes.io/worker= -o jsonpath="{{range .items[*]}}{{.metadata.name}}{{'\\n'}}{{end}}" |  xargs -I{{}} {self.__cli} debug node/{{}} -- chroot /host sh -c "mkdir -p /etc/kata-containers; cp /usr/share/kata-containers/defaults/configuration.toml /etc/kata-containers/; sed -i 's/thread-pool-size=1/thread-pool-size={thread_pool_size}/' /etc/kata-containers/configuration.toml" """)

    def delete_kata_threads_pool(self):
        """
        This method deletes kata thread-pool-size from every worker node
        @return:
        """
        self.run(fr"""{self.__cli} get nodes -l node-role.kubernetes.io/worker= -o jsonpath="{{range .items[*]}}{{.metadata.name}}{{'\\n'}}{{end}}" |  xargs -I{{}} {self.__cli} debug node/{{}} -- chroot /host sh -c "rm -f /etc/kata-containers/configuration.toml" """)

    @typechecked
    def populate_additional_template_variables(self, env: dict):
        """
        This method populates any additional variables needed for setup templates
        """
        if self.__kata_csv:  # custom kata version
            env['kata_csv'] = self.__kata_csv
        else:  # latest kata version
            env['kata_csv'] = self._get_kata_csv()
        env['kata_channel'] = self._get_kata_channel()
        env['kata_catalog_source'] = self._get_kata_catalog_source()
        env['kata_namespace'] = self._get_kata_namespace()

    def is_cnv_installed(self):
        """
        This method checks if cnv operator is installed
        :return:
        """
        verify_cmd = f"{self.__cli} get csv -n openshift-cnv -o jsonpath='{{.items[0].status.phase}}'"
        if 'Succeeded' in self.run(verify_cmd):
            return True
        return False

    def is_odf_installed(self):
        """
        This method checks if odf operator is installed
        :return:
        """
        verify_cmd = f"{self.__cli} get csv -n openshift-storage -o jsonpath='{{.items[0].status.phase}}'"
        if 'Succeeded' in self.run(verify_cmd):
            return True
        return False

    def check_dv_status(self,
                        status: str,
                        namespace: str = environment_variables.environment_variables_dict['namespace']):
        """
        This method checks dv status
        :return:
        """
        namespace = f'-n {namespace}' if namespace else ''
        verify_cmd = f"{self.__cli} get dv {namespace} -o jsonpath='{{.items[].status.phase}}'"
        if status in self.run(verify_cmd):
            return True
        return False

    @typechecked
    @logger_time_stamp
    def wait_for_dv_status(self,
                           status: str = 'Succeeded',
                           timeout: int = int(environment_variables.environment_variables_dict['timeout'])):
        """
        This method waits for methods status
        @return: True/ False if reach to status
        """
        current_wait_time = 0
        while timeout <= 0 or current_wait_time <= timeout:
            if self.check_dv_status(status=status):
                return True
            # sleep for x seconds
            time.sleep(OC.SLEEP_TIME)
            current_wait_time += OC.SLEEP_TIME
        raise DVStatusTimeout(status=status)

    def verify_odf_installation(self, namespace: str = 'openshift-storage'):
        """
        This method verifies ODF installation
        :return: True ODF passed, False failed
        """
        self.run(
            f"{self.__cli} patch storagecluster ocs-storagecluster -n {namespace} --type json --patch '[{{ \"op\": \"replace\", \"path\": \"/spec/enableCephTools\", \"value\": true }}]'")
        self.wait_for_patch(pod_name='rook-ceph-tools', label='app=rook-ceph-tools', label_uuid=False, namespace=namespace)
        health_check = self.run(
            f"{self.__cli} -n {namespace} rsh {self._get_pod_name(pod_name='rook-ceph-tools', namespace=namespace)} ceph health")
        return 'HEALTH_OK' == health_check.strip()

    def get_odf_disk_count(self):
        """
        This method returns odf disk count
        :return:
        """
        if self.is_odf_installed():
            return int(self.run(f"{self.__cli} get --no-headers pod -n openshift-storage | grep osd | grep -cv prepare"))

    def is_kata_installed(self):
        """
        This method checks if kata operator is installed
        :return:
        """
        verify_cmd = f"{self.__cli} get csv -n openshift-sandboxed-containers-operator -o jsonpath='{{.items[0].status.phase}}'"
        if 'Succeeded' in self.run(verify_cmd):
            return True
        return False

    def get_master_nodes(self):
        """
        This method returns master nodes
        :return:
        """
        return self.run(fr""" {self.__cli} get nodes -l node-role.kubernetes.io/master= -o jsonpath="{{range .items[*]}}{{.metadata.name}}{{'\n'}}{{end}}" """)

    def get_worker_nodes(self):
        """
        This method returns worker nodes
        :return:
        """
        return self.run(fr""" {self.__cli} get nodes -l node-role.kubernetes.io/worker= -o jsonpath="{{range .items[*]}}{{.metadata.name}}{{'\n'}}{{end}}" """)

    def delete_available_released_pv(self):
        """
        This method deletes available or released pv because that avoid launching new pv
        """
        pv_status_list = self.run(fr"{self.__cli} get pv -o jsonpath={{..status.phase}}").split()
        for ind, pv_status in enumerate(pv_status_list):
            if pv_status == 'Available' or pv_status == 'Released':
                available_pv = self.run(fr"{self.__cli} get pv -o jsonpath={{.items[{ind}].metadata.name}}")
                logger.info(f'Delete {pv_status} pv {available_pv}')
                self.run(fr"{self.__cli} delete localvolume -n openshift-local-storage local-disks")
                self.run(fr"{self.__cli} delete pv {available_pv}")

    def clear_node_caches(self):
        """
        This method clears the node's buffer cache
        """
        return self.run(fr""" {self.__cli} get nodes -l node-role.kubernetes.io/worker= -o jsonpath="{{range .items[*]}}{{.metadata.name}}{{'\n'}}{{end}}" |  xargs -I{{}} {self.__cli} debug node/{{}} -- chroot /host sh -c "sync; echo 3 > /proc/sys/vm/drop_caches" """)

    def __get_short_uuid(self, workload: str):
        """
        This method returns uuid
        :return:
        """
        long_uuid = self.get_long_uuid(workload=workload)
        uuids = long_uuid.split('-')
        short_uuid = uuids[0]
        return short_uuid

    def get_num_active_nodes(self):
        """
        This method returns the number of active nodes
        :return:
        """
        # count the number of active master/worker nodes
        if self.get_worker_nodes():
            count_nodes = len(self.get_worker_nodes().split())
        else:
            count_nodes = len(self.get_master_nodes().split())
        return count_nodes

    @typechecked
    @logger_time_stamp
    def apply_security_privileged(self, namespace: str = environment_variables.environment_variables_dict['namespace']):
        """
        This method applies security privileged for namespace
        @param namespace:
        @return:
        """
        return self.run(f"{self.__cli} create serviceaccount -n {namespace} {namespace}; "
                        f"{self.__cli} adm policy add-scc-to-user -n {namespace} privileged -z {namespace}")

    @typechecked
    @logger_time_stamp
    def create_async(self, yaml: str, is_check: bool = False):
        """
        This method creates yaml in async
        @param yaml:
        @param is_check:
        :return:
        """
        if os.path.isfile(yaml):
            return self.run(f'{self.__cli} create -f {yaml}', is_check=is_check)
        else:
            raise YAMLNotExist(yaml)

    @typechecked
    @logger_time_stamp
    def apply_async(self, yaml: str, is_check: bool = False):
        """
        This method creates yaml in async
        @param yaml:
        @param is_check:
        :return:
        """
        if os.path.isfile(yaml):
            return self.run(f'{self.__cli} apply -f {yaml}', is_check=is_check)
        else:
            raise YAMLNotExist(yaml)

    @typechecked
    @logger_time_stamp
    def delete_async(self, yaml: str):
        """
        This method deletes yaml in async
        :param yaml:
        :return:
        """
        if os.path.isfile(yaml):
            return self.run(f'{self.__cli} delete -f {yaml}')
        else:
            raise YAMLNotExist(yaml)

    @typechecked
    def _get_pod_name(self, pod_name: str,
                      namespace: str = environment_variables.environment_variables_dict['namespace']):
        """
        This method returns pod name if exist or raise error
        :param pod_name:
        :param namespace:
        :return:
        """
        try:
            namespace = f'-n {namespace}' if namespace else ''
            return self.run(f'{self.__cli} get {namespace} pods -o name | grep {pod_name}')
        except Exception as err:
            raise PodNameNotExist(pod_name=pod_name)

    def pod_exists(self, pod_name: str,
                   namespace: str = environment_variables.environment_variables_dict['namespace']):
        """
        This method returns True if exist or False if not
        :param pod_name:
        :param namespace:
        :return:
        """
        namespace = f'-n {namespace}' if namespace else ''
        result = self.run(f'{self.__cli} get {namespace} pods -o name | grep {pod_name}')
        if pod_name in result:
            return True
        else:
            return False

    def pod_label_exists(self, label_name: str,
                         namespace: str = environment_variables.environment_variables_dict['namespace']):
        """
        This method returns True if pod exist or not by label
        :param label_name:
        :param namespace:
        :return:
        """
        namespace = f'-n {namespace}' if namespace else ''
        result = self.run(f"{self.__cli} get {namespace} pod -l={label_name} -o jsonpath='{{.items}}'")
        if result != '[]':
            return True
        else:
            return False

    @typechecked()
    def get_long_uuid(self, workload: str):
        """
        This method returns uuid
        :return:
        """
        long_uuid = self.run(
            f"{self.__cli} -n {environment_variables.environment_variables_dict['namespace']} get benchmark/{workload} -o jsonpath={{.status.uuid}}")
        return long_uuid

    def get_ocp_major_version(self):
        """
        This method returns ocp major version
        @return:
        """
        return int(self.get_ocp_server_version().split('.')[0])

    def get_ocp_minor_version(self):
        """
        This method returns ocp minor version
        @return:
        """
        return int(self.get_ocp_server_version().split('.')[1])

    def get_prom_token(self):
        """
        This method returns prometheus token
        :return:
        """
        # OCP 4.10 and below
        if self.get_ocp_major_version() <= 4 and self.get_ocp_minor_version() <= 10:
            prom_token = self.run(f"{self.__cli} -n openshift-monitoring sa get-token prometheus-k8s")
        else:
            prom_token = self.run(f"{self.__cli} sa new-token -n openshift-monitoring prometheus-k8s 2>/dev/null")
        return prom_token

    def collect_events(self):
        """
        This method collects event log
        :return: output_filename
        """
        output_filename = os.path.join(self._run_artifacts, f'events.log')
        self.run(f"{self.__cli} get events -A > '{output_filename}' ")
        return output_filename

    @logger_time_stamp
    def login(self):
        """
        This method logs in to the cluster
        :return:
        """
        try:
            if self.__kubeadmin_password and self.__kubeadmin_password != '':
                self.run(f'{self.__cli} login -u kubeadmin -p {self.__kubeadmin_password}', is_check=True)
        except Exception as err:
            raise LoginFailed
        return True

    @typechecked
    @logger_time_stamp
    def get_pod(self, label: str, database: str = '', namespace: str = environment_variables.environment_variables_dict['namespace']):
        """
        This method gets pods according to label
        :param label:
        :param database:
        :param namespace:
        :return:
        """
        if database:
            return self.run(
                f"{self.__cli} get pods -n '{database}-db'" + " --no-headers | awk '{ print $1; }' | grep " + database,
                is_check=True).rstrip().decode('ascii')
        else:
            namespace = f'-n {namespace}' if namespace else ''
            return self.run(f"{self.__cli} get pods {namespace} --no-headers | awk '{{ print $1; }}' | grep -w '{label}'", is_check=True).rstrip().decode('ascii')

    @typechecked
    @logger_time_stamp
    def save_pod_log(self, pod_name: str, database: str = '', log_type: str = ''):
        """
        This method saves pod log in log_path
        :param pod_name: pod name with uuid
        :param database: database
        :param log_type: log type extension
        :return: output_filename
        """
        if log_type:
            output_filename = os.path.join(self._run_artifacts, f'{pod_name}{log_type}')
        else:
            output_filename = os.path.join(self._run_artifacts, pod_name)
        if database:
            self.run(f"{self.__cli} logs -n '{database}-db' {pod_name} > '{output_filename}' ")
        # manager logs of benchmark-controller-manager
        elif 'benchmark-controller-manager' in pod_name:
            self.run(f"{self.__cli} logs -n {environment_variables.environment_variables_dict['namespace']} {pod_name} manager > '{output_filename}' ")
        else:
            self.run(f"{self.__cli} logs -n {environment_variables.environment_variables_dict['namespace']} {pod_name} > '{output_filename}' ")
        return output_filename

    def describe_pod(self, pod_name: str, namespace: str = ''):
        """
        This method describes pod into log
        :param pod_name: pod name with uuid
        :param namespace: namespace
        :return: output_filename
        """
        output_filename = os.path.join(self._run_artifacts, f'describe-{pod_name}')
        self.run(f"{self.__cli} describe pod -n {namespace} {pod_name} > '{output_filename}' ")
        return output_filename

    @logger_time_stamp
    def get_pods(self):
        """
        This method retrieves information on benchmark-runner pods in oc get pod format
        :return:
        """
        return self.run(f'{self.__cli} get pods', is_check=True)

    @typechecked
    @logger_time_stamp
    def wait_for_pod_create(self, pod_name: str,
                            namespace: str = environment_variables.environment_variables_dict['namespace'],
                            timeout: int = int(environment_variables.environment_variables_dict['timeout'])):
        """
        This method waits till pod name is creating or throw exception after timeout
        :param namespace:
        :param pod_name:
        :param timeout:
        :return: True if getting pod name or raise PodNameError
        """
        current_wait_time = 0
        while timeout <= 0 or current_wait_time <= timeout:
            if self.pod_exists(pod_name=pod_name, namespace=namespace):
                self.describe_pod(pod_name=pod_name, namespace=namespace)
                return True
            # sleep for x seconds
            time.sleep(OC.SLEEP_TIME)
            current_wait_time += OC.SLEEP_TIME
        self.describe_pod(pod_name=pod_name, namespace=namespace)
        raise PodNotCreateTimeout(pod_name)

    @typechecked
    @logger_time_stamp
    def wait_for_pod_terminate(self, pod_name: str,
                               namespace: str = environment_variables.environment_variables_dict['namespace'],
                               timeout: int = int(environment_variables.environment_variables_dict['timeout'])):
        """
        This method waits till pod name is terminating or throw exception after timeout
        :param namespace:
        :param pod_name:
        :param timeout:
        :return: True if pod name terminated or raise
        """
        current_wait_time = 0
        while timeout <= 0 or current_wait_time <= timeout:
            if not self.pod_exists(pod_name=pod_name, namespace=namespace):
                return True
            # sleep for x seconds
            time.sleep(OC.SLEEP_TIME)
            current_wait_time += OC.SLEEP_TIME
        raise PodTerminateTimeout(pod_name)

    @typechecked
    @logger_time_stamp
    def create_pod_sync(self, yaml: str, pod_name: str,
                        namespace: str = environment_variables.environment_variables_dict['namespace'],
                        timeout: int = int(environment_variables.environment_variables_dict['timeout'])):
        """
        This method creates pod yaml in async
        :param namespace:
        :param timeout:
        :param pod_name:
        :param yaml:
        :return:
        """
        self.create_async(yaml)
        return self.wait_for_pod_create(pod_name=pod_name, namespace=namespace, timeout=timeout)

    @typechecked
    @logger_time_stamp
    def delete_pod_sync(self, yaml: str, pod_name: str,
                        namespace: str = environment_variables.environment_variables_dict['namespace'],
                        timeout: int = int(environment_variables.environment_variables_dict['timeout'])):
        """
        This method deletes pod yaml in async, only if exist and return false if not exist
        :param namespace:
        :param timeout:
        :param pod_name:
        :param yaml:
        :return:
        """
        if self.pod_exists(pod_name, namespace):
            self.delete_async(yaml)
            return self.wait_for_pod_terminate(pod_name=pod_name, namespace=namespace, timeout=timeout)
        else:
            return False

    @typechecked
    def delete_namespace(self, namespace: str = environment_variables.environment_variables_dict['namespace']):
        """
        This method deletes namespace
        :param namespace:
        :return:
        """
        self.run(f'{self.__cli} delete ns {namespace}')

    @typechecked
    @logger_time_stamp
    def wait_for_initialized(self, label: str, workload: str = '', status: str = 'Initialized', label_uuid: bool = True,
                             namespace: str = environment_variables.environment_variables_dict['namespace'],
                             timeout: int = SHORT_TIMEOUT):
        """
        This method waits to pod to be initialized
        :param namespace:
        :param label:
        :param status:
        :param label_uuid: need to get uuid from label (benchmark-operator)
        :param timeout:
        :param workload:
        :return:
        """
        try:
            namespace = f'-n {namespace}' if namespace else ''
            if label_uuid:
                result = self.run(
                    f"{self.__cli} {namespace} wait --for=condition={status} pod -l {label}-{self.__get_short_uuid(workload=workload)} --timeout={timeout}s",
                    is_check=True)
            else:
                return self.run(
                    f"{self.__cli} {namespace} wait --for=condition={status} pod -l {label} --timeout={timeout}s",
                    is_check=True)
            if 'met' in result.decode("utf-8"):
                return True
        except Exception as err:
            if 'vm' in workload:
                raise VMNotInitializedTimeout(workload=workload)
            else:
                raise PodNotInitializedTimeout(workload=workload)

    @typechecked
    @logger_time_stamp
    def wait_for_upgrade_version(self, operator: str, upgrade_version: str,
                                 namespace: str = environment_variables.environment_variables_dict['namespace'],
                                 timeout: int = int(
                                     environment_variables.environment_variables_dict['timeout'])) -> bool:
        """
        This method waits until all operators' CSVs reach the expected upgrade version.

        :param operator: The operator for which the upgrade is being monitored.
        :param upgrade_version: The expected version that all CSVs should reach.
        :param namespace: The namespace in which the operator CSVs are located.
        :param timeout: The maximum time to wait for the upgrade to complete.
        :return: True if all CSVs reach the expected version, or raise OperatorUpgradeTimeout.
        """
        current_wait_time = 0

        while timeout <= 0 or current_wait_time <= timeout:
            upgrade_versions = self.run(
                f"{self.__cli} get csv -n {namespace} -o custom-columns=:.spec.version --no-headers").splitlines()
            count_upgrade_version = sum(1 for actual_upgrade_version in upgrade_versions if
                                        '.'.join(actual_upgrade_version.split('.')[0:2]) == upgrade_version)

            if len(upgrade_versions) == count_upgrade_version:
                return True

            # Sleep for a predefined time before checking again
            time.sleep(OC.SLEEP_TIME)
            current_wait_time += OC.SLEEP_TIME

        raise OperatorUpgradeTimeout(operator=operator, version=upgrade_version, namespace=namespace)

    @typechecked
    @logger_time_stamp
    def wait_for_csv(self, operator: str, csv_num: int = 1,
                     namespace: str = environment_variables.environment_variables_dict['namespace'],
                     timeout: int = int(environment_variables.environment_variables_dict['timeout'])) -> str:
        """
        This method waits until the required number of CSVs are created or throws an exception after a timeout.

        :param operator: The operator for which the CSVs are being monitored.
        :param csv_num: The required number of CSVs, default is 1.
        :param namespace: The namespace in which the operator CSVs are located.
        :param timeout: The maximum time to wait for the CSVs to be created.
        :return: CSV names if the required number of CSVs are found, or raises CSVNotCreateTimeout.
        """
        current_wait_time = 0

        while timeout <= 0 or current_wait_time <= timeout:
            csv_names = self.run(f"{self.__cli} get csv -n {namespace} -o jsonpath={{$.items[*].metadata.name}}")
            if csv_names and len(csv_names.split()) >= csv_num:
                return csv_names

            # Sleep for a predefined time before checking again
            time.sleep(OC.SLEEP_TIME)
            current_wait_time += OC.SLEEP_TIME

        raise CSVNotCreateTimeout(operator, namespace)

    @typechecked
    @logger_time_stamp
    def wait_for_patch(self, pod_name: str, label: str, label_uuid: bool, namespace: str, timeout: int = SHORT_TIMEOUT):
        """
        This method waits for patch, needs to wait that pod is created and then wait for ready
        @param pod_name:
        @param label:
        @param label_uuid:
        @param namespace:
        @param timeout:
        @return:
        """
        current_wait_time = 0
        while timeout <= 0 or current_wait_time <= timeout:
            if self._get_pod_name(pod_name=pod_name, namespace=namespace) and self.wait_for_ready(label=label, label_uuid=label_uuid, namespace=namespace):
                return True
            # sleep for x seconds
            time.sleep(OC.SLEEP_TIME)
            current_wait_time += OC.SLEEP_TIME
        raise PodNotReadyTimeout(label)

    @typechecked
    @logger_time_stamp
    def wait_for_ready(self, label: str, run_type: str = 'pod', workload: str = '', status: str = 'ready', label_uuid: bool = True,
                       namespace: str = environment_variables.environment_variables_dict['namespace'],
                       timeout: int = SHORT_TIMEOUT):
        """
        This method waits for the pod/vm to be ready
        :param label:
        :param run_type:pod or vm
        :param workload:
        :param namespace:
        :param status:
        :param label_uuid: need to get uuid from label (benchmark-operator)
        :param timeout:
        :return:
        """
        try:
            namespace = f'-n {namespace}' if namespace else ''
            if label_uuid:
                result = self.run(
                    f"{self.__cli} {namespace} wait --for=condition={status} {run_type} -l {label}-{self.__get_short_uuid(workload=workload)} --timeout={timeout}s",
                    is_check=True)
            else:
                result = self.run(
                    f"{self.__cli} {namespace} wait --for=condition={status} {run_type} -l {label} --timeout={timeout}s",
                    is_check=True)
            if 'met' in result.decode("utf-8"):
                return True
        except Exception as err:
            if 'vm' in workload:
                raise VMNotReadyTimeout(workload=workload)
            else:
                raise PodNotReadyTimeout(workload=workload)

    @typechecked
    def exec(self, command: str, pod_name: str, namespace: str = environment_variables.environment_variables_dict['namespace'], container: str = ''):
        """
        This method executes a command within a specified pod and optional container and returns the output
        :param command:
        :param pod_name:
        :param namespace:
        :param container:
        :return:
        """
        try:
            namespace = f'-n {namespace}' if namespace else ''
            container = f'-c {container}' if container else ''
            return self.run(f'{self.__cli} exec {namespace} {pod_name} {container} -- {command}')
        except Exception as err:
            raise ExecFailed(command, pod_name, err)

    @typechecked
    def terminate_pod_sync(self, pod_name: str,
                           namespace: str = environment_variables.environment_variables_dict['namespace'],
                           timeout: int = int(environment_variables.environment_variables_dict['timeout'])):
        """
        This method deletes a pod based on name and namespace only
        :param pod_name:
        :param namespace:
        :param timeout:
        :return:
        """
        if self.pod_exists(pod_name, namespace):
            try:
                namespace = f'-n {namespace}' if namespace else ''
                self.run(f'{self.__cli} delete pod {namespace} {pod_name} timeout={timeout}')
            except Exception as err:
                raise PodTerminateTimeout(pod_name)

    @typechecked
    @logger_time_stamp
    def wait_for_pod_ready(self, pod_name: str,
                           namespace: str = environment_variables.environment_variables_dict['namespace'],
                           timeout: int = int(environment_variables.environment_variables_dict['timeout'])):
        """
        This method waits for a pod to be ready and running
        :param pod_name:
        :param namespace:
        :param timeout:
        :return: True if getting pod name or raise PodNameError
        """
        self.wait_for_pod_create(pod_name=pod_name, namespace=namespace, timeout=timeout)
        current_wait_time = 0
        namespace = f'-n {namespace}' if namespace else ''
        while timeout <= 0 or current_wait_time <= timeout:
            answer = self.run(f'{self.__cli} get pod {namespace} {pod_name} --no-headers -ocustom-columns=Status:status.phase 2>/dev/null')
            if answer == 'Running':
                return
            elif answer == 'Error':
                raise PodFailed(pod_name)
            time.sleep(OC.SLEEP_TIME)
            current_wait_time += OC.SLEEP_TIME
        raise PodNotReadyTimeout(pod_name)

    @typechecked
    @logger_time_stamp
    def wait_for_pod_completed(self, label: str, workload: str = '', label_uuid: bool = True, job: bool = True,
                               namespace: str = environment_variables.environment_variables_dict['namespace'],
                               timeout: int = int(environment_variables.environment_variables_dict['timeout'])):
        """
        This method waits for a pod to be completed
        :param workload:
        :param label:
        :param label_uuid: need to get uuid from label (benchmark-operator)
        :param job: kind is job instead of pod
        :param timeout:
        :param namespace:
        :return:
        """
        try:
            current_wait_time = 0
            namespace = f'-n {namespace}' if namespace else ''
            while current_wait_time <= timeout:
                if label_uuid and job:
                    result = self.run(
                        f"{self.__cli} {namespace} wait --for=condition=complete -l {label}-{self.__get_short_uuid(workload=workload)} jobs --timeout={OC.SHORT_TIMEOUT}s")
                    if 'met' in result:
                        return True
                    result = self.run(
                        f"{self.__cli} {namespace} wait --for=condition=failed -l {label}-{self.__get_short_uuid(workload=workload)} jobs --timeout={OC.SLEEP_TIME}s")
                    if 'met' in result:
                        return False
                if not job:
                    result = self.run(f"{self.__cli} get pod -l {label}" + " -n benchmark-runner --no-headers | awk '{ print $3; }'")
                    if 'Completed' in result:
                        return True
            # sleep for x seconds
            time.sleep(OC.SLEEP_TIME)
            current_wait_time += OC.SLEEP_TIME
        except Exception as err:
            raise PodNotCompletedTimeout(workload=workload)

    def describe_vmi(self, vm_name: str, namespace: str):
        """
        This method describes vmi into log
        :param vm_name: vm name with uuid
        :param namespace: namespace
        :return: output_filename
        """
        output_filename = os.path.join(self._run_artifacts, f'describe-{vm_name}')
        self.run(f"{self.__cli} describe vmi -n {namespace} {vm_name} > '{output_filename}' ")
        return output_filename

    @typechecked
    def _get_vm_name(self, vm_name: str, namespace: str = environment_variables.environment_variables_dict['namespace']):
        """
        This method returns VM name if the vm exists, otherwise raise VMNameNotExist exception
        :param vm_name:
        :param namespace:
        :return: VM name or raise VMNameNotExist
        """
        try:
            namespace = f'-n {namespace}' if namespace else ''
            return self.run(f'{self.__cli} get {namespace} vmi -o name | grep {vm_name}', is_check=True)
        except Exception as err:
            raise VMNameNotExist(vm_name=vm_name)

    @typechecked
    def _get_all_vm_names(self, namespace: str = environment_variables.environment_variables_dict['namespace']):
        """
        This method returns a list of VM names in the given namespace.

        :param namespace: str, the namespace to look for VMs in. Defaults to the namespace in environment_variables_dict.
        :return: list of VM names or an empty list if an error occurs
        """
        namespace_option = f'-n {namespace}' if namespace else ''
        command = f"{self.__cli} get {namespace_option} vm -o jsonpath='{{.items[*].metadata.name}}'"
        try:
            vm_names = self.run(command)
            return vm_names.split() if vm_names else []
        except Exception:
            return []

    @typechecked
    def vm_exists(self, vm_name: str, namespace: str = environment_variables.environment_variables_dict['namespace']):
        """
        This method returns True or False if vm name exist
        :param vm_name:
        :param namespace:
        :return: True or False
        """
        namespace = f'-n {namespace}' if namespace else ''
        result = self.run(f'{self.__cli} get {namespace} vmi -o name | grep {vm_name}')
        if vm_name in result:
            return True
        else:
            return False

    @typechecked
    @logger_time_stamp
    def get_vm(self, label: str = '', namespace: str = environment_variables.environment_variables_dict['namespace']):
        """
        This method gets vm according to label
        :param label:
        :param namespace:
        :return:
        """
        if label:
            namespace = f'-n {namespace}' if namespace else ''
            return self.run(
                cmd=f"{self.__cli} get vmi {namespace} --no-headers | awk '{{ print $1; }}' | grep -w '{label}'", is_check=True).rstrip().decode('ascii')
        else:
            return self.run(f'{self.__cli} get vmi', is_check=True)

    @logger_time_stamp
    def __verify_vm_log_complete(self, vm_name: str, timeout: int = int(environment_variables.environment_variables_dict['timeout'])):
        """
        This method verifies that vm log is complete
        :param vm_name: vm name with uuid
        :return:
        """
        output_filename = f"{self._run_artifacts}/{vm_name}"
        current_wait_time = 0
        while timeout <= 0 or current_wait_time <= timeout:
            with open(output_filename) as file:
                if '-----END SSH HOST KEY KEYS-----' in file.read():
                    return True
            # sleep for x seconds
            time.sleep(OC.SLEEP_TIME)
            current_wait_time += OC.SLEEP_TIME
        raise VMNotCompletedTimeout(vm_name)

    @typechecked
    @logger_time_stamp
    def get_exposed_vm_port(self, vm_name: str, namespace: str = environment_variables.environment_variables_dict['namespace']):
        """
        The method gets exposed vm port
        @param vm_name:
        @return:
        """
        namespace = f'-n {namespace}' if namespace else ''
        return self.run(cmd=f"{self.__cli} get service {vm_name} {namespace} -o jsonpath={{.spec.ports[].nodePort}}")

    @logger_time_stamp
    def get_nodes_addresses(self):
        """
        This method returns dictionary of nodes and corresponding IP addresses, e.g. {node1:ip1, node2:ip2, node3:ip3 }
        :return:
        """
        node_ips = self.run(f"{self.__cli} get node -o jsonpath='{{$.items[*].status.addresses[*].address}}'")
        node_ips_list = node_ips.split()
        return dict([(k, v) for k, v in zip(node_ips_list[1::2], node_ips_list[::2])])

    @logger_time_stamp
    def wait_for_vm_status(self, vm_name: str = '', status: VMStatus = VMStatus.Stopped,
                           namespace: str = environment_variables.environment_variables_dict['namespace'],
                           timeout: int = SHORT_TIMEOUT):
        """
        This method waits for VM to reach the specified status
        :param vm_name:
        :param status: Stopped, Starting, Running
        :param namespace:
        :param timeout:
        :return:
        """
        current_wait_time = 0
        namespace = f'-n {namespace}' if namespace else ''
        while timeout <= 0 or current_wait_time <= timeout:
            check_vm_status = f"{self.__cli} get vm {vm_name} {namespace} -o jsonpath={{.status.printableStatus}}"
            result = self.run(check_vm_status)
            if result == status.name:
                return True
            # sleep for x seconds
            time.sleep(OC.SLEEP_TIME)
            current_wait_time += OC.SLEEP_TIME
        raise VMStateTimeout(vm_name=vm_name, state=status)

    def wait_for_vm_ssh(self, vm_name: str = '', node_ip: str = '', vm_node_port: str = '',
                        timeout: int = SHORT_TIMEOUT):
        """
        This method waits for VM to be accessible via ssh login
        :param vm_name:
        :param node_ip:
        :param vm_node_port:
        :param timeout:
        :return:
        """
        current_wait_time = 0
        while timeout <= 0 or current_wait_time <= timeout:
            check_vm_ssh = f"""if [ "$(ssh -o 'BatchMode=yes' -o ConnectTimeout=1 root@{node_ip} -p {vm_node_port} 2>&1|egrep 'denied|verification failed')" ]; then echo 'True'; else echo 'False'; fi"""
            result = self.run(check_vm_ssh)
            if result == 'True':
                return True
            # sleep for x seconds
            time.sleep(OC.SLEEP_TIME)
            current_wait_time += OC.SLEEP_TIME
        raise VMStateTimeout(vm_name=vm_name, state='ssh')

    @logger_time_stamp
    def get_vm_node(self, vm_name: str, namespace: str = environment_variables.environment_variables_dict['namespace']):
        """
        This method gets vm node
        :param vm_name:
        :param namespace:
        :return:
        """
        namespace = f'-n {namespace}' if namespace else ''
        command = f"{self.__cli} get vmi {vm_name} {namespace} -o jsonpath={{.metadata.labels.'kubevirt\\.io/nodeName'}}"

        try:
            result = self.run(command)
            if result and "NotFound" not in result:
                return result.strip()
            return None
        except Exception as e:
            # Log the exception details if necessary
            print(f"Error occurred: {e}")
            return None

    @typechecked
    @logger_time_stamp
    def wait_for_vm_create(self, vm_name: str,
                           namespace: str = environment_variables.environment_variables_dict['namespace'],
                           timeout: int = int(environment_variables.environment_variables_dict['timeout'])):
        """
        This method waits till vm name is creating or throw exception after timeout
        :param vm_name:
        :param namespace:
        :param timeout:
        :return: True if getting pod name or raise PodNameError
        """
        current_wait_time = 0
        while timeout <= 0 or current_wait_time <= timeout:
            if self.vm_exists(vm_name=vm_name, namespace=namespace):
                self.describe_vmi(vm_name=vm_name, namespace=namespace)
                return True
            # sleep for x seconds
            time.sleep(OC.SLEEP_TIME)
            current_wait_time += OC.SLEEP_TIME
        self.describe_vmi(vm_name=vm_name, namespace=namespace)
        raise VMNotCreateTimeout(vm_name)

    @typechecked
    @logger_time_stamp
    def create_vm_sync(self, yaml: str, vm_name: str,
                       namespace: str = environment_variables.environment_variables_dict['namespace'],
                       timeout: int = int(environment_variables.environment_variables_dict['timeout'])):
        """
        This method creates vm synchronously
        :param timeout:
        :param vm_name:
        :param yaml:
        :return:
        """
        self.create_async(yaml)
        return self.wait_for_vm_create(vm_name=vm_name, namespace=namespace, timeout=timeout)

    @typechecked
    @logger_time_stamp
    def delete_vm_sync(self, yaml: str, vm_name: str,
                       namespace: str = environment_variables.environment_variables_dict['namespace'],
                       timeout: int = int(environment_variables.environment_variables_dict['timeout'])):
        """
        This method deletes specified VM synchronously; return False if it does not exist
        :param namespace:
        :param timeout:
        :param vm_name:
        :param yaml:
        :return: return False if vm does not exist
        """
        if self.vm_exists(vm_name=vm_name, namespace=namespace):
            self.delete_async(yaml)
            return self.wait_for_vm_delete(vm_name=vm_name, namespace=namespace, timeout=timeout)
        else:
            return False

    @logger_time_stamp
    def delete_all_vms(self, namespace: str = environment_variables.environment_variables_dict['namespace']):
        """
        This method deletes all vms
        :return:
        """
        namespace = f'-n {namespace}' if namespace else ''
        self.run(f"{self.__cli} delete vm --all --grace-period 0 {namespace}")

    @logger_time_stamp
    def wait_for_vm_completed(self, workload: str = '', vm_name: str = '',
                              namespace: str = environment_variables.environment_variables_dict['namespace'],
                              timeout: int = int(environment_variables.environment_variables_dict['timeout'])):
        """
        This method waits for the VM to complete
        :param vm_name: The vm name
        :param workload:
        :param namespace:
        :return: True or raise error VMNotCompletedTimeout
        """
        current_wait_time = 0
        namespace = f'-n {namespace}' if namespace else ''
        while timeout <= 0 or current_wait_time <= timeout:
            if self.run(
                    f"{self.__cli} {namespace} get benchmark {workload} -o jsonpath={{.status.complete}}") == 'true':
                return True
            # sleep for x seconds
            time.sleep(OC.SLEEP_TIME)
            current_wait_time += OC.SLEEP_TIME
        raise VMNotCompletedTimeout(workload=workload)

    @typechecked
    @logger_time_stamp
    def wait_for_vm_delete(self, vm_name: str,
                           namespace: str = environment_variables.environment_variables_dict['namespace'],
                           timeout: int = int(environment_variables.environment_variables_dict['timeout'])):
        """
        This method waits till VM delete; raise VMTerminateTimeout if it does not delete within the specified timeout
        :param vm_name:
        :param namespace:
        :param timeout:
        :return: True if pod name terminated or raise
        """
        current_wait_time = 0
        while timeout <= 0 or current_wait_time <= timeout:
            if not self.vm_exists(vm_name=vm_name, namespace=namespace):
                return True
            # sleep for x seconds
            time.sleep(OC.SLEEP_TIME)
            current_wait_time += OC.SLEEP_TIME
        raise VMDeleteTimeout(vm_name)

    @logger_time_stamp
    def wait_for_vm_log_completed(self, vm_name: str = '', end_stamp: str = '', output_filename: str = '',
                                  timeout: int = int(environment_variables.environment_variables_dict['timeout']),
                                  sleep_time: int = 30):
        """
        This method wait to vm to be completed by end sign string
        :param vm_name:
        :param end_stamp: end of run stamp
        :param output_filename:
        :param timeout:
        :param sleep_time:
        :return: True or through timeout
        """
        if not output_filename:
            output_filename = os.path.join(self._run_artifacts, vm_name)
        current_wait_time = 0
        # wait initialize time till vm log create
        time.sleep(sleep_time)
        while timeout <= 0 or current_wait_time <= timeout:
            with open(output_filename, 'r') as file:
                data = file.read()
                if end_stamp in data:
                    return True
                else:
                    # sleep for x seconds
                    time.sleep(sleep_time)
                    current_wait_time += sleep_time
        raise VMNotCompletedTimeout(workload=vm_name)

    @logger_time_stamp
    def extract_vm_results(self, vm_name: str = '', start_stamp: str = '', end_stamp: str = '', output_filename: str = ''):
        """
        This method extracts vm results from vm output log
        :param vm_name:
        :param start_stamp: start of run stamp
        :param end_stamp: end of run stamp
        :param output_filename:
        :return:
        """
        data_index = 1
        title_index = 2
        if not output_filename:
            output_filename = os.path.join(self._run_artifacts, vm_name)
        results_list = []
        with open(output_filename) as infile:
            # VM logs results start
            start = False
            for line in infile:
                if start_stamp in line or 'podman' in line and not start:
                    start = True
                    continue
                elif end_stamp in line:
                    return results_list
                elif start:
                    # Filter 'cloud-init' and CSV lines only
                    if 'cloud-init' in line and ',' in line:
                        if vm_name in line:
                            # filter the title, placed after the second :
                            results_list.append(line.strip().split(':')[title_index:])
                        else:
                            # filter the data, placed after the first :
                            results_list.append(line.strip().split(':')[data_index:])
        return results_list




import os

from benchmark_runner.common.oc.oc import OC
from benchmark_runner.common.logger.logger_time_stamp import logger_time_stamp, logger
from benchmark_runner.common.ocp_resources.create_ocp_resource_operations import CreateOCPResourceOperations


class CreateLSO(CreateOCPResourceOperations):
    """
    This class is created Local Storage operator
    """
    def __init__(self, oc: OC, path: str, resource_list: list):
        super().__init__(oc)
        self.__oc = oc
        self.__path = path
        self.__resource_list = resource_list

    @logger_time_stamp
    def create_lso(self, upgrade_version: str):
        """
        This method creates lso operator
        :return:
        """
        if upgrade_version:
            self.__oc.apply_async(yaml=os.path.join(self.__path, '03_subscription.yaml'))
            logger.info(f'Wait till LSO upgrade to version: {upgrade_version}')
            self.verify_csv_installation(namespace='openshift-local-storage', operator='lso', upgrade_version=upgrade_version)
        else:
            for resource in self.__resource_list:
                logger.info(f'run {resource}')
                self.__oc.create_async(yaml=os.path.join(self.__path, resource))

            # verify once after create all resource files
            self.wait_for_ocp_resource_create(operator='lso',
                                              verify_cmd="oc -n openshift-local-storage wait deployment/local-storage-operator --for=condition=Available",
                                              status='deployment.apps/local-storage-operator condition met')
            return True



import ast
import os
from jinja2 import Template

from benchmark_runner.common.oc.oc import OC
from benchmark_runner.main.environment_variables import environment_variables
from benchmark_runner.common.ocp_resources.create_lso import CreateLSO
from benchmark_runner.common.ocp_resources.create_odf import CreateODF
from benchmark_runner.common.ocp_resources.create_kata import CreateKata
from benchmark_runner.common.ocp_resources.create_cnv import CreateCNV
from benchmark_runner.common.ocp_resources.create_custom import CreateCustom
from benchmark_runner.common.ocp_resources.migrate_infra import MigrateInfra


class CreateOCPResource:
    """
    This class create ocp resources
    """
    def __init__(self):
        self.__dir_path = f'{os.path.dirname(os.path.realpath(__file__))}'
        self.__environment_variables_dict = environment_variables.environment_variables_dict
        self.__oc = OC(kubeadmin_password=self.__environment_variables_dict.get('kubeadmin_password', ''))
        self.__oc.login()
        self.__oc.populate_additional_template_variables(self.__environment_variables_dict)
        self.__worker_disk_prefix = self.__environment_variables_dict.get('worker_disk_prefix', '')
        self.__worker_disk_ids = self.__environment_variables_dict.get('worker_disk_ids', '')
        if self.__worker_disk_ids:
            if self.__worker_disk_ids:
                # Solved GitHub Actions issue that env variable detect as string instead of dict/ list -skip for Jenkins
                if self.__environment_variables_dict.get('github_token', ''):
                    self.__worker_disk_ids = self.__worker_disk_ids.replace('"', '')
                self.__worker_disk_ids = ast.literal_eval(self.__worker_disk_ids)

    @staticmethod
    def __get_yaml_files(path: str, extensions: list = ['.yaml', '.sh']):
        file_list = []
        for file in os.listdir(path):
            for extension in extensions:
                if file.endswith(extension):
                    file_list.append(file)
        return file_list

    def remove_resource_files(self, path: str, extensions: list = ['.yaml', '.sh']):
        """
        This method remove yamls files in path
        :param path: The path
        :param extensions: File type to delete
        :return:
        """
        file_list = self.__get_yaml_files(path=path, extensions=extensions)
        for file in file_list:
            os.remove(os.path.join(path, file))

    def get_sorted_resources(self, resource: str):
        """
        This method update environment variable inside yaml file
        :param resource:
        :return:
        """
        for resource_file in os.listdir(os.path.join(self.__dir_path, resource, 'template')):
            with open(os.path.join(self.__dir_path, resource, 'template', resource_file)) as f:
                template_str = f.read()
            tm = Template(template_str, keep_trailing_newline=True)
            data = tm.render(self.__environment_variables_dict)
            resource_file = resource_file.replace('_template', '')
            with open(os.path.join(self.__dir_path, resource, resource_file), 'w') as f:
                f.write(data)
        resource_file_list = self.__get_yaml_files(path=os.path.join(self.__dir_path, resource))
        resource_file_list = sorted(resource_file_list, key=lambda x: os.path.splitext(x)[0])
        return resource_file_list

    def create_resource(self, resource: str, upgrade_version: str):
        """
        This method create resource with verification
        :param resource:
        :param upgrade_version:
        :return:
        """
        # remove resource files
        self.remove_resource_files(path=os.path.join(self.__dir_path, resource))
        resource_files = self.get_sorted_resources(resource=resource)
        if 'lso' == resource:
            create_lso = CreateLSO(self.__oc, path=os.path.join(self.__dir_path, resource), resource_list=resource_files)
            create_lso.create_lso(upgrade_version)
        elif 'odf' == resource:
            create_odf = CreateODF(self.__oc, path=os.path.join(self.__dir_path, resource), resource_list=resource_files, worker_disk_ids=self.__worker_disk_ids, worker_disk_prefix=self.__worker_disk_prefix)
            create_odf.create_odf(upgrade_version)
        elif 'kata' == resource:
            create_kata = CreateKata(self.__oc, path=os.path.join(self.__dir_path, resource), resource_list=resource_files)
            create_kata.create_kata()
        elif 'cnv' == resource:
            create_cnv = CreateCNV(self.__oc, path=os.path.join(self.__dir_path, resource), resource_list=resource_files)
            create_cnv.create_cnv()
        elif 'infra' == resource:
            create_infra = MigrateInfra(self.__oc, path=os.path.join(self.__dir_path, resource), resource_list=resource_files)
            create_infra.migrate_infra()
        elif 'custom' == resource:
            create_custom = CreateCustom(self.__oc, path=os.path.join(self.__dir_path, resource), resource_list=resource_files)
            create_custom.create_custom()
            # remove resource files
        self.remove_resource_files(path=os.path.join(self.__dir_path, resource))




import sys
import os
import logging
from benchmark_runner.main.environment_variables import environment_variables

logger_category_name = 'benchmark_runner.benchmark'
logger = logging.getLogger(logger_category_name)  # instantiating a logger
handler = logging.StreamHandler(sys.stdout)
environment_variables_dict = environment_variables.environment_variables_dict
# log_path = os.getcwd()
# logs folder
run_artifacts = environment_variables_dict.get('run_artifacts', '')
run_artifacts_path = environment_variables_dict.get('run_artifacts_path', '')
if not os.path.isdir(run_artifacts_path):
    os.makedirs(run_artifacts, exist_ok=True)
    os.mkdir(run_artifacts_path)
fileHandler = logging.FileHandler(filename=f'{run_artifacts_path}/benchmark_runner.log', mode='w+')
logger.addHandler(handler)
logger.addHandler(fileHandler)


def get_log_path():
    """
    This method return log path
    :return:
    """
    return f'{run_artifacts_path}/benchmark_runner.log'




import os
import yaml
import datetime
import tarfile
import shutil
from typeguard import typechecked

from benchmark_runner.common.logger.logger_time_stamp import logger_time_stamp, logger
from benchmark_runner.common.oc.oc import OC
from benchmark_runner.common.virtctl.virtctl import Virtctl
from benchmark_runner.common.template_operations.template_operations import TemplateOperations
from benchmark_runner.common.elasticsearch.elasticsearch_operations import ElasticSearchOperations
from benchmark_runner.common.ssh.ssh import SSH
from benchmark_runner.benchmark_operator.benchmark_operator_exceptions import ODFNotInstalled, CNVNotInstalled, KataNotInstalled, EmptyLSOPath
from benchmark_runner.main.environment_variables import environment_variables
from benchmark_runner.common.clouds.shared.s3.s3_operations import S3Operations
from benchmark_runner.common.prometheus.prometheus_snapshot import PrometheusSnapshot
from benchmark_runner.common.prometheus.prometheus_snapshot_exceptions import PrometheusSnapshotError
from benchmark_runner.common.prometheus.prometheus_metrics_operations import PrometheusMetricsOperation


class BenchmarkOperatorWorkloadsOperations:
    """
    This class contains benchmark-operator workload operations
    """
    def __init__(self):
        # environment variables
        self._environment_variables_dict = environment_variables.environment_variables_dict
        self._workload = self._environment_variables_dict.get('workload', '')
        self._workload_name = self._workload.replace('_', '-')
        self._time_stamp_format = self._environment_variables_dict.get('time_stamp_format', '')
        self._runner_version = self._environment_variables_dict.get('build_version', '')
        self._run_type = self._environment_variables_dict.get('run_type', '')
        self._workloads_odf_pvc = list(self._environment_variables_dict.get('workloads_odf_pvc', ''))
        self._odf_pvc = self._environment_variables_dict.get('odf_pvc', '')
        self._storage_type = self._environment_variables_dict.get('storage_type', '')
        self._system_metrics = self._environment_variables_dict.get('system_metrics', '')
        self._elasticsearch = self._environment_variables_dict.get('elasticsearch', '')
        self._run_artifacts = self._environment_variables_dict.get('run_artifacts', '')
        self._run_artifacts_path = self._environment_variables_dict.get('run_artifacts_path', '')
        self._date_key = self._environment_variables_dict.get('date_key', '')
        self._key = self._environment_variables_dict.get('key', '')
        self._endpoint_url = self._environment_variables_dict.get('endpoint_url', '')
        self._save_artifacts_local = self._environment_variables_dict.get('save_artifacts_local', '')
        self._enable_prometheus_snapshot = self._environment_variables_dict.get('enable_prometheus_snapshot', False)
        self._kubeadmin_password = self._environment_variables_dict.get('kubeadmin_password', '')
        self._dir_path = f'{os.path.dirname(os.path.realpath(__file__))}'
        self._pin_node_benchmark_operator = self._environment_variables_dict.get('pin_node_benchmark_operator', '')
        self._pin_node1 = self._environment_variables_dict.get('pin_node1', '')
        self._pin_node2 = self._environment_variables_dict.get('pin_node2', '')
        self._es_host = self._environment_variables_dict.get('elasticsearch', '')
        self._es_port = self._environment_variables_dict.get('elasticsearch_port', '')
        self._es_user = self._environment_variables_dict.get('elasticsearch_user', '')
        self._es_password = self._environment_variables_dict.get('elasticsearch_password', '')
        self._timeout = int(self._environment_variables_dict.get('timeout', ''))
        self._uuid = self._environment_variables_dict.get('uuid', '')
        self._es_url_protocol = self._environment_variables_dict['elasticsearch_url_protocol']
        self._ssh = SSH()
        # Elasticsearch connection
        if self._es_host and self._es_port:
            self.__es_operations = ElasticSearchOperations(es_host=self._es_host,
                                                           es_port=self._es_port,
                                                           es_user=self._es_user,
                                                           es_password=self._es_password,
                                                           es_url_protocol=self._es_url_protocol,
                                                           timeout=self._timeout)
        # Generate template class
        self._template = TemplateOperations(workload=self._workload)
        # set oc login
        self._oc = self.set_login(kubeadmin_password=self._kubeadmin_password)
        self._virtctl = Virtctl()
        # PrometheusSnapshot
        if self._enable_prometheus_snapshot:
            self._snapshot = PrometheusSnapshot(oc=self._oc, artifacts_path=self._run_artifacts_path, verbose=True)
        self._prometheus_metrics_operation = PrometheusMetricsOperation()
        # Extract lso id for LSO workload
        if '_lso' in self._environment_variables_dict.get('workload'):
            # Update lso_disk_id only if both worker_disk_ids and a free disk exist
            if self._environment_variables_dict.get('worker_disk_ids', '') and self._oc.get_free_disk_id(
                    node=self._environment_variables_dict['lso_node']):
                self._lso_disk_id = self._oc.get_free_disk_id(node=self._environment_variables_dict['lso_node'])
                self._environment_variables_dict['lso_disk_id'] = self._lso_disk_id
            else:
                self._lso_disk_id = self._environment_variables_dict.get('lso_disk_id', '')
        else:
            self._lso_disk_id = None

    def set_login(self, kubeadmin_password: str = ''):
        """
        This method sets oc login
        :param kubeadmin_password:
        :return:
        """
        self._oc = OC(kubeadmin_password=kubeadmin_password)
        self._oc.login()
        return self._oc

    @logger_time_stamp
    def update_node_selector(self, runner_path: str = environment_variables.environment_variables_dict['runner_path'], yaml_path: str = '', pin_node: str = ''):
        """
        This method updates node selector in yaml
        @return:
        """
        data = []
        # Read YAML file and inject node selector in the right place
        with open(os.path.join(runner_path, yaml_path), 'r') as stream:
            try:
                documents = yaml.safe_load_all(stream)
                for doc in documents:
                    if doc.get('spec'):
                        doc['spec']['template']['spec']['nodeSelector'] = {
                            'kubernetes.io/hostname': f"{self._environment_variables_dict.get(pin_node, '')}"}
                    data.append(doc)
            except yaml.YAMLError as exc:
                logger.error(exc)

        # Write YAML file
        with open(os.path.join(self._environment_variables_dict.get('runner_path', ''),
                               'benchmark-operator/config/manager/manager.yaml'), 'w', encoding='utf8') as outfile:
            yaml.safe_dump_all(data, outfile, default_flow_style=False, allow_unicode=True)

    @typechecked()
    @logger_time_stamp
    def make_deploy_benchmark_controller_manager(self, runner_path: str = environment_variables.environment_variables_dict['runner_path']):
        """
        This method deploys the benchmark operator
        :return:
        """
        benchmark_operator_path = 'benchmark-operator'
        current_dir = os.getcwd()
        os.chdir(os.path.join(runner_path, benchmark_operator_path))
        # Patch for custom image: export IMG=quay.io/user/benchmark-operator:latest;
        self._ssh.run('make deploy')
        self._oc.wait_for_pod_create(pod_name='benchmark-controller-manager')
        self._oc.wait_for_initialized(label='control-plane=controller-manager', label_uuid=False)
        self._oc.wait_for_ready(label='control-plane=controller-manager', label_uuid=False)
        os.chdir(current_dir)

    @typechecked()
    @logger_time_stamp
    def make_undeploy_benchmark_controller_manager(self, runner_path: str = environment_variables.environment_variables_dict['runner_path']):
        """
       This method removes a benchmark-operator deployment
        :return:
        """
        benchmark_operator_path = 'benchmark-operator'
        current_dir = os.getcwd()
        os.chdir(os.path.join(runner_path, benchmark_operator_path))
        self._ssh.run('make undeploy')
        self._oc.wait_for_pod_terminate(pod_name='benchmark-controller-manager')
        os.chdir(current_dir)

    @logger_time_stamp
    def make_undeploy_benchmark_controller_manager_if_exist(self, runner_path: str = environment_variables.environment_variables_dict['runner_path']):
        """
        This method undeploys benchmark controller manager, if it exists
        @return:
        """
        # delete benchmark-operator pod if exist
        if self._oc.pod_exists(pod_name='benchmark-controller-manager'):
            logger.info('make undeploy benchmark operator running pod')
            self.make_undeploy_benchmark_controller_manager(runner_path=runner_path)

    @logger_time_stamp
    def login(self):
        """
        This method logs in to the cluster
        """
        self._oc.login()

    @logger_time_stamp
    def tear_down_pod_after_error(self, yaml: str, pod_name: str):
        """
        This method tears down pod in case of error
        @param yaml:
        @param pod_name:
        @return:
        """
        self._oc.delete_pod_sync(yaml=yaml, pod_name=pod_name)
        self.delete_all()
        self.make_undeploy_benchmark_controller_manager(runner_path=environment_variables.environment_variables_dict['runner_path'])

    @logger_time_stamp
    def tear_down_vm_after_error(self, yaml: str, vm_name: str):
        """
        This method tears down vm in case of error
        @param yaml:
        @param vm_name:
        """
        self._oc.delete_vm_sync(yaml=yaml, vm_name=vm_name)
        self.delete_all()
        self.make_undeploy_benchmark_controller_manager(runner_path=environment_variables.environment_variables_dict['runner_path'])

    @logger_time_stamp
    def system_metrics_collector(self, workload: str, es_fetch_min_time: int = None):
        """
        This method runs system metrics collector
        @param workload: the workload
        @param es_fetch_min_time:
        :return:
        """
        if self._run_type == 'test_ci':
            es_index = 'system-metrics-test-ci'
        else:
            es_index = 'system-metrics'
        self._oc.wait_for_pod_create(pod_name='system-metrics-collector')
        self._oc.wait_for_initialized(label='app=system-metrics-collector', workload=workload)
        self._oc.wait_for_pod_completed(label='app=system-metrics-collector', workload=workload)
        # verify that data upload to elastic search
        if self._es_host:
            # es_fetch_min_time: if we need to fetch more time
            if es_fetch_min_time:
                self._verify_elasticsearch_data_uploaded(index=es_index, uuid=self._oc.get_long_uuid(workload=workload), workload='system-metrics', fast_check=True, es_fetch_min_time=es_fetch_min_time)
            else:
                self._verify_elasticsearch_data_uploaded(index=es_index, uuid=self._oc.get_long_uuid(workload=workload), workload='system-metrics', fast_check=True)

    def __get_metadata(self, kind: str = None, database: str = None, status: str = None, run_artifacts_url: str = None, uuid: str = None, prometheus_result: dict = None) -> dict:
        """
        This method returns metadata for a run, optionally filtered by runtime kind and database
        @param kind: optionally: pod, vm, or kata
        @param database: optional:mssql, postgres or mariadb
        @param status:
        @param run_artifacts_url:
        :return:
        """
        date_format = '%Y_%m_%d'
        metadata = {'ocp_version': self._oc.get_ocp_server_version(),
                    'cnv_version': self._oc.get_cnv_version(),
                    'kata_version': self._oc.get_kata_operator_version(),
                    'kata_rpm_version': self._oc.get_kata_rpm_version(node=self._pin_node1),
                    'odf_version': self._oc.get_odf_version(),
                    'runner_version': self._runner_version,
                    'version': int(self._runner_version.split('.')[-1]),
                    'ci_date': datetime.datetime.now().strftime(date_format),
                    'pin_node_benchmark_operator': self._pin_node_benchmark_operator,
                    'pin_node1': self._pin_node1,
                    'pin_node2': self._pin_node2,
                    'storage_type': self._storage_type}
        if kind:
            metadata.update({'kind': kind})
        if status:
            metadata.update({'run_status': status})
        if run_artifacts_url:
            metadata.update({'run_artifacts_url': run_artifacts_url})
        if database:
            metadata.update({'vm_os_version': 'centos-stream8'})
        else:
            metadata.update({'vm_os_version': 'fedora34'})
        if uuid:
            metadata.update({'uuid': uuid})
        if prometheus_result:
            metadata.update(prometheus_result)
        # for hammerdb
        if database == 'mssql':
            metadata.update({'db_version': 2019})
        elif database == 'postgres':
            metadata.update({'db_version': 10})
        elif database == 'mariadb':
            metadata.update({'db_version': 10.3})
        return metadata

    @logger_time_stamp
    def _update_elasticsearch_index(self, index: str, id: str, kind: str, status: str, run_artifacts_url: str, database: str = '', prometheus_result: dict = None):
        """
        This method updates elasticsearch id
        :param index:
        :param id:
        :param kind:
        :param database:
        :param status:
        :param run_artifacts_url:
        :return:
        """
        self.__es_operations.update_elasticsearch_index(index=index, id=id, metadata=self.__get_metadata(kind=kind, database=database, status=status, run_artifacts_url=run_artifacts_url, prometheus_result=prometheus_result))

    def _upload_to_elasticsearch(self, index: str, kind: str, status: str, run_artifacts_url: str, database: str = '', uuid: str = ''):
        """
        This method uploads data to elasticsearch
        :param index:
        :param kind:
        :param status:
        :param run_artifacts_url:
        :param database:
        :return:
        """
        self.__es_operations.upload_to_elasticsearch(index=index, data=self.__get_metadata(kind=kind, database=database, status=status, run_artifacts_url=run_artifacts_url, uuid=uuid))

    def _verify_elasticsearch_data_uploaded(self, index: str, uuid: str, workload: str = '', fast_check: bool = False, timeout: int = None, es_fetch_min_time: int = None):
        """
        This method verifies that elasticsearch data uploaded
        :param index:
        :param uuid:
        :param workload:
        :param fast_check:
        :param timeout:
        :return: ids
        """
        if workload:
            return self.__es_operations.verify_elasticsearch_data_uploaded(index=index, uuid=uuid, workload=workload, fast_check=fast_check, timeout=timeout, es_fetch_min_time=es_fetch_min_time)
        else:
            return self.__es_operations.verify_elasticsearch_data_uploaded(index=index, uuid=uuid, workload=self._workload_name, fast_check=fast_check, timeout=timeout, es_fetch_min_time=es_fetch_min_time)

    def _upload_workload_to_elasticsearch(self, index: str, kind: str, status: str, result: dict = None):
        """
        This method uploads to elasticsearch
        :param index:
        :param kind:
        :param status:
        :param result:
        :return:
        """
        self.__es_operations.upload_to_elasticsearch(index=index, data=self.__get_metadata(kind=kind, status=status, result=result))

    def _verify_elasticsearch_workload_data_uploaded(self, index: str, uuid: str):
        """
        This method verifies that elasticsearch data was uploaded
        :param index:
        :param uuid:
        :return:
        """
        self.__es_operations.verify_elasticsearch_data_uploaded(index=index, uuid=uuid)

    def _create_vm_log(self, labels: list) -> str:
        """
        This method sets vm log per workload
        :param labels: list of labels
        :return: vm_name
        """
        vm_name = ''
        for label in labels:
            vm_name = self._oc.get_vm(label=label)
            self._virtctl.save_vm_log(vm_name=vm_name)
        return vm_name

    def _create_pod_log(self, label: str = '', database: str = '') -> str:
        """
        This method creates pod log per workload
        :param label:pod label
        :param database:
        :return:
        """
        pod_name = self._oc.get_pod(label=label, database=database)
        if database:
            self._oc.save_pod_log(pod_name=pod_name, database=database)
        else:
            self._oc.save_pod_log(pod_name=pod_name)

    def _get_run_artifacts_hierarchy(self, workload_name: str = '', is_file: bool = False):
        """
        This method returns log hierarchy
        :param workload_name: workload name
        :param is_file: is file name
        :return:
        """
        key = self._key
        run_type = self._run_type.replace('_', '-')
        date_key = self._date_key
        if workload_name:
            workload_key = workload_name.split('-')[0]
            if is_file:
                return os.path.join(key, run_type, date_key, workload_key, workload_name)
            else:
                return os.path.join(key, run_type, date_key, workload_key)
        return os.path.join(key, run_type, date_key)

    def _create_run_artifacts(self, workload: str = '', database: str = '', labels: list = [], pod: bool = True):
        """
        This method creates pod logs of benchmark-controller-manager, system-metrics and workload pod
        :param workload: workload name
        :param database: database name
        :param pod: False in case of vm
        :param labels: list of labels of pod names - using it when it different from workload name
        :return: run artifacts url
        """
        self._create_pod_log(label='benchmark-controller-manager')
        if environment_variables.environment_variables_dict['system_metrics']:
            self._create_pod_log(label='system-metrics')
        # for vm call to create_vm_log
        if pod:
            # workload that contains 2 pods
            if labels:
                for label in labels:
                    self._create_pod_log(label=label)
            else:
                self._create_pod_log(label=workload)
            if database:
                self._create_pod_log(database=database)
        workload_name = self._workload_name
        return os.path.join(self._environment_variables_dict.get('run_artifacts_url', ''), f'{self._get_run_artifacts_hierarchy(workload_name=workload_name, is_file=True)}-{self._time_stamp_format}.tar.gz')

    def __make_run_artifacts_tarfile(self, workload: str):
        """
        This method compresses the log file and returns the compressed path
        :return:
        """
        tar_run_artifacts_path = f"{self._run_artifacts_path}.tar.gz"
        with tarfile.open(tar_run_artifacts_path, mode='w:gz') as archive:
            archive.add(self._run_artifacts_path, arcname=f'{workload}-{self._time_stamp_format}', recursive=True)
        return tar_run_artifacts_path

    @logger_time_stamp
    def delete_local_artifacts(self):
        """
        This method deletes local artifacts
        :return:
        """
        workload = self._workload_name
        tar_run_artifacts_path = self.__make_run_artifacts_tarfile(workload)
        # remove local run artifacts workload folder
        # verify that its not empty path
        if len(self._run_artifacts_path) > 3 and self._run_artifacts_path != '/' and self._run_artifacts_path and tar_run_artifacts_path and os.path.isfile(tar_run_artifacts_path) and not self._save_artifacts_local:
            # remove run_artifacts_path
            shutil.rmtree(path=self._run_artifacts_path)
            # remove tar.gz file
            os.remove(path=tar_run_artifacts_path)

    @logger_time_stamp
    def upload_run_artifacts_to_s3(self):
        """
        This method uploads log to s3
        :return:
        """
        workload = self._workload_name
        tar_run_artifacts_path = self.__make_run_artifacts_tarfile(workload)
        run_artifacts_hierarchy = self._get_run_artifacts_hierarchy(workload_name=workload)
        # Upload when endpoint_url is not None
        s3operations = S3Operations()
        # change workload to key convention
        upload_file = f"{workload}-{self._time_stamp_format}.tar.gz"
        s3operations.upload_file(file_name_path=tar_run_artifacts_path,
                                 bucket=self._environment_variables_dict.get('bucket', ''),
                                 key=run_artifacts_hierarchy,
                                 upload_file=upload_file)

    @logger_time_stamp
    def start_prometheus(self):
        """
        This method starts collection of Prometheus snapshot
        :return:
        """
        if self._enable_prometheus_snapshot:
            try:
                self._snapshot.prepare_for_snapshot()
            except PrometheusSnapshotError as err:
                raise PrometheusSnapshotError(err)
            except Exception as err:
                raise err

    @logger_time_stamp
    def end_prometheus(self):
        """
        This method retrieves the Prometheus snapshot
        :return:
        """
        if self._enable_prometheus_snapshot:
            try:
                self._snapshot.retrieve_snapshot()
            except PrometheusSnapshotError as err:
                raise PrometheusSnapshotError(err)
            except Exception as err:
                raise err

    @logger_time_stamp
    def odf_workload_verification(self):
        """
        This method verifies whether the ODF operator is installed for ODF workloads and raises an error if it is missing.
        :return:
        """
        workload_name = self._workload.split('_')
        if workload_name[0] in self._workloads_odf_pvc:
            if not self._oc.is_odf_installed():
                raise ODFNotInstalled(workload=self._workload)

    @logger_time_stamp
    def verify_lso(self):
        """
        This method verifies that lso path exist
        :return:
        """
        if not self._lso_disk_id:
            raise EmptyLSOPath()

    @logger_time_stamp
    def delete_all(self):
        """
        This method deletes all pod or unbound pv in namespace
        :return:
        """
        # make undeploy benchmark controller manager if exist
        self.make_undeploy_benchmark_controller_manager_if_exist(runner_path=environment_variables.environment_variables_dict['runner_path'])
        if 'hammerdb' in self._workload:
            self._oc.delete_namespace(namespace=f"{self._workload.split('_')[2]}-db")
        if self._storage_type == 'lso':
            self._oc.delete_available_released_pv()
            self._oc.remove_lso_path()

    @logger_time_stamp
    def clear_nodes_cache(self):
        """
        This method clears nodes cache
        """
        self._oc.clear_node_caches()

    def initialize_workload(self):
        """
        This method includes all the initialization of workload
        :return:
        """
        # Verify that CNV operator in installed for CNV workloads
        if '_vm' in self._workload and not self._oc.is_cnv_installed():
            raise CNVNotInstalled(workload=self._workload)
        # Verify that Kata operator in installed for kata workloads
        if '_kata' in self._workload and not self._oc.is_kata_installed():
            raise KataNotInstalled(workload=self._workload)
        self.delete_all()
        self.clear_nodes_cache()
        if self._odf_pvc:
            self.odf_workload_verification()
        if 'lso' in self._workload:
            self.verify_lso()
        # make deploy benchmark controller manager
        self.make_deploy_benchmark_controller_manager(runner_path=environment_variables.environment_variables_dict['runner_path'])
        self._template.generate_yamls()
        if self._enable_prometheus_snapshot:
            self.start_prometheus()

    def finalize_workload(self):
        """
        This method includes all the finalization of workload
        :return:
        """
        self._oc.collect_events()
        if self._enable_prometheus_snapshot:
            self.end_prometheus()
        if self._endpoint_url:
            self.upload_run_artifacts_to_s3()
        if not self._save_artifacts_local:
            self.delete_local_artifacts()
        self.delete_all()




from codecs import open
from os import path
from setuptools import setup, find_packages


__version__ = '1.0.639'  # update also .bumpversion.cfg


here = path.abspath(path.dirname(__file__))


if path.isfile(path.join(here, 'README.md')):
    with open(path.join(here, 'README.md'), encoding='utf-8') as f:
        long_description = f.read()
else:
    long_description = ""

setup(
    name='benchmark-runner',
    version=__version__,
    description='Benchmark Runner Tool',
    long_description=long_description,
    long_description_content_type='text/markdown',
    author='Red Hat',
    author_email='ebattat@redhat.com',
    url='https://github.com/redhat-performance/benchmark-runner',
    license="Apache License 2.0",
    classifiers=[
        'License :: OSI Approved :: Apache Software License',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'Programming Language :: Python :: 3.12',
    ],

    zip_safe=False,

    # Find all packages (__init__.py)
    packages=find_packages(include=['benchmark_runner', 'benchmark_runner.*']),

    install_requires=[
        'attrs==21.4.0',  # readthedocs
        'azure==4.0.0',
        'boto3==1.33.13',  # s3
        'botocore==1.33.13',  # s3
        'cryptography==42.0.4',  # for paramiko
        'elasticsearch==7.16.1',
        'elasticsearch_dsl==7.4.0',  # for deep search
        'google-auth==2.30.0',  # for Google Drive
        'google-auth-httplib2==0.2.0',  # for Google Drive
        'google-auth-oauthlib==1.2.0',  # for Google Drive
        'google-api-python-client==2.135.0',  # for Google Drive
        'ipywidgets==8.0.6',  # for jupyterlab widgets
        'jinja2==3.1.4',  # for yaml templates and df.style
        'myst-parser==1.0.0',  # readthedocs
        'openshift-client==1.0.17',  # clusterbuster && prometheus metrics
        'prometheus-api-client==0.5.1',  # clusterbuster && prometheus metrics
        'pandas',  # required latest
        'paramiko==3.4.0',
        'PyGitHub==1.55',  # update secrets
        'PyYAML==6.0.1',
        'setuptools',  # for python3.12
        'sphinx==5.0.0',  # readthedocs
        'sphinx-rtd-theme==1.0.0',  # readthedocs
        'tenacity==8.0.1',  # retry decorator
        'tqdm==4.66.3',  # for jupyterlab download file
        'typeguard==2.12.1',
        'typing==3.7.4.3',
        # must add new package inside requirements.txt
    ],

    setup_requires=['pytest', 'pytest-runner', 'wheel', 'coverage'],

    include_package_data=True,

    # dependency_links=[]
)




import ast
import os
import time
import datetime
import tarfile
import shutil
from csv import DictReader

from benchmark_runner.common.logger.logger_time_stamp import logger_time_stamp
from benchmark_runner.workloads.workloads_exceptions import ODFNotInstalled, CNVNotInstalled, KataNotInstalled, MissingScaleNodes, MissingRedis
from benchmark_runner.common.oc.oc import OC
from benchmark_runner.common.virtctl.virtctl import Virtctl
from benchmark_runner.common.elasticsearch.elasticsearch_operations import ElasticSearchOperations
from benchmark_runner.main.environment_variables import environment_variables
from benchmark_runner.common.clouds.shared.s3.s3_operations import S3Operations
from benchmark_runner.common.prometheus.prometheus_snapshot import PrometheusSnapshot
from benchmark_runner.common.prometheus.prometheus_snapshot_exceptions import PrometheusSnapshotError
from benchmark_runner.common.template_operations.template_operations import TemplateOperations
from benchmark_runner.common.clouds.IBM.ibm_operations import IBMOperations
from benchmark_runner.common.prometheus.prometheus_metrics_operations import PrometheusMetricsOperation


class WorkloadsOperations:
    oc = None
    MILLISECONDS = 1000
    REPEAT_TIMES = 3
    SLEEP_TIME = 3
    """
    This class contains workloads operations
    """
    def __init__(self):
        # environment variables
        self._environment_variables_dict = environment_variables.environment_variables_dict
        self._workload = self._environment_variables_dict.get('workload', '')
        self._build_version = self._environment_variables_dict.get('build_version', '')
        self._workloads_odf_pvc = list(self._environment_variables_dict.get('workloads_odf_pvc', ''))
        self._odf_pvc = self._environment_variables_dict.get('odf_pvc', True)
        self._kubeadmin_password = self._environment_variables_dict.get('kubeadmin_password', '')
        self._run_type = self._environment_variables_dict.get('run_type', '')
        self._trunc_uuid = self._environment_variables_dict.get('trunc_uuid', '')
        self._uuid = self._environment_variables_dict.get('uuid', '')
        self._date_key = self._environment_variables_dict.get('date_key', '')
        self._key = self._environment_variables_dict.get('key', '')
        self._time_stamp_format = self._environment_variables_dict.get('time_stamp_format', '')
        self._endpoint_url = self._environment_variables_dict.get('endpoint_url', '')
        self._run_artifacts_path = self._environment_variables_dict.get('run_artifacts_path', '')
        self._save_artifacts_local = self._environment_variables_dict.get('save_artifacts_local', '')
        self._enable_prometheus_snapshot = self._environment_variables_dict.get('enable_prometheus_snapshot', False)
        self._run_artifacts_url = self._environment_variables_dict.get('run_artifacts_url', '')
        self._pin_node1 = self._environment_variables_dict.get('pin_node1', '')
        self._pin_node2 = self._environment_variables_dict.get('pin_node2', '')
        self._es_host = self._environment_variables_dict.get('elasticsearch', '')
        self._es_port = self._environment_variables_dict.get('elasticsearch_port', '')
        self._es_user = self._environment_variables_dict.get('elasticsearch_user', '')
        self._es_password = self._environment_variables_dict.get('elasticsearch_password', '')
        self._es_url_protocol = self._environment_variables_dict['elasticsearch_url_protocol']
        self._scale = self._environment_variables_dict.get('scale', '')
        self._redis = self._environment_variables_dict.get('redis', '')
        self._threads_limit = self._environment_variables_dict.get('threads_limit', '')
        self._kata_thread_pool_size = self._environment_variables_dict.get('kata_thread_pool_size', '')
        if self._scale:
            self._scale = int(self._scale)
            self._scale_nodes = self._environment_variables_dict.get('scale_nodes', '')
            self._redis = self._environment_variables_dict.get('redis', '')
            if not self._scale_nodes:
                raise MissingScaleNodes()
            if not self._redis and 'vdbench' in self._workload:
                raise MissingRedis()
            self._scale_node_list = ast.literal_eval(self._scale_nodes)
            if self._threads_limit:
                self._threads_limit = int(self._threads_limit)
            else:
                self._threads_limit = self._scale * len(self._scale_node_list)
            self._bulk_sleep_time = int(self._environment_variables_dict.get('bulk_sleep_time', ''))
        else:
            self._scale_node_list = []
        self._timeout = int(self._environment_variables_dict.get('timeout', ''))
        # Elasticsearch connection
        if self._es_host and self._es_port:
            self._es_operations = ElasticSearchOperations(es_host=self._es_host,
                                                           es_port=self._es_port,
                                                           es_user=self._es_user,
                                                           es_password=self._es_password,
                                                           es_url_protocol=self._es_url_protocol,
                                                           timeout=self._timeout)
        # Generate templates class
        self._template = TemplateOperations(workload=self._workload)

        # set oc login
        if WorkloadsOperations.oc is None:
            WorkloadsOperations.oc = self.set_login(kubeadmin_password=self._kubeadmin_password)
        self._oc = WorkloadsOperations.oc
        self._virtctl = Virtctl()

        # Prometheus Snapshot
        if self._enable_prometheus_snapshot:
            self._snapshot = PrometheusSnapshot(oc=self._oc, artifacts_path=self._run_artifacts_path, verbose=True)
        self._prometheus_snap_interval = self._environment_variables_dict.get('prometheus_snap_interval', '')
        self._prometheus_metrics_operation = PrometheusMetricsOperation()
        self._windows_url = self._environment_variables_dict.get('windows_url', '')
        self._delete_all = self._environment_variables_dict.get('delete_all', '')
        self._verification_only = self._environment_variables_dict.get('verification_only', '')
        if self._windows_url:
            file_name = os.path.basename(self._windows_url)
            self._windows_os = os.path.splitext(file_name)[0]

    def _get_workload_file_name(self, workload):
        """
        This method returns workload name
        :return:
        """
        if self._scale:
            return f'{workload}-scale-{self._time_stamp_format}'
        else:
            return f'{workload}-{self._time_stamp_format}'

    def set_login(self, kubeadmin_password: str = ''):
        """
        This method sets login
        :param kubeadmin_password:
        :return: oc instance
        """
        self._oc = OC(kubeadmin_password=kubeadmin_password)
        self._oc.login()
        return self._oc

    @logger_time_stamp
    def delete_all(self):
        """
        This method deletes all resources in namespace
        :return:
        """
        self._oc.delete_namespace()

    @logger_time_stamp
    def start_prometheus(self):
        """
        This method starts collection of Prometheus snapshot
        :return:
        """
        if self._enable_prometheus_snapshot:
            try:
                self._snapshot.prepare_for_snapshot()
            except PrometheusSnapshotError as err:
                raise PrometheusSnapshotError(err)
            except Exception as err:
                raise err

    @logger_time_stamp
    def end_prometheus(self):
        """
        This method retrieves the Prometheus snapshot
        :return:
        """
        if self._enable_prometheus_snapshot:
            try:
                self._snapshot.retrieve_snapshot()
            except PrometheusSnapshotError as err:
                raise PrometheusSnapshotError(err)
            except Exception as err:
                raise err

    @logger_time_stamp
    def odf_workload_verification(self):
        """
        This method verifies whether the ODF operator is installed for ODF workloads and raises an error if it is missing.
        :return:
        """
        workload_name = self._workload.split('_')
        if workload_name[0] in self._workloads_odf_pvc:
            if not self._oc.is_odf_installed():
                raise ODFNotInstalled(workload=self._workload)

    def _create_vm_log(self, labels: list) -> str:
        """
        This method sets vm log per workload
        :param labels: list of labels
        :return: vm_name
        """
        vm_name = ''
        for label in labels:
            vm_name = self._oc.get_vm(label=label)
            self._virtctl.save_vm_log(vm_name=vm_name)
        return vm_name

    def _create_pod_log(self, pod: str = '', log_type: str = ''):
        """
        This method creates pod log per workload
        :param pod: pod name
        :return: save_pod_log file
        """
        pod_name = self._oc.get_pod(label=pod)
        return self._oc.save_pod_log(pod_name=pod_name, log_type=log_type)

    def _get_run_artifacts_hierarchy(self, workload_name: str = '', is_file: bool = False):
        """
        This method returns log hierarchy
        :param workload_name: workload name
        :param is_file: is file name
        :return:
        """
        key = self._key
        run_type = self._run_type.replace('_', '-')
        date_key = self._date_key
        if workload_name:
            workload_key = workload_name.split('-')[0]
            if is_file:
                return os.path.join(key, run_type, date_key, workload_key, workload_name)
            else:
                return os.path.join(key, run_type, date_key, workload_key)
        return os.path.join(key, run_type, date_key)

    @staticmethod
    def __is_float(value) -> bool:
        """
        This method checks if value is float
        :param value: The value to check
        :return: True if the value is a float, False when cannot be converted to a float
        """
        try:
            float(value)
            return True
        except (ValueError, TypeError):
            return False

    def _create_scale_logs(self):
        """
        The method creates scale logs
        :return:
        """
        self._create_pod_log(pod='state-signals-exporter', log_type='.log')
        self._create_pod_log(pod='redis-master', log_type='.log')

    def _create_pod_run_artifacts(self, pod_name: str, log_type: str):
        """
        This method creates pod run artifacts
        :param pod_name: pod name
        :param log_type: log type extension
        :return: run results list of dicts
        """
        result_list = []
        pod_log_file = self._create_pod_log(pod=pod_name, log_type=log_type)
        workload_name = self._environment_variables_dict.get('workload', '').replace('_', '-')
        # csv to dictionary
        the_reader = DictReader(open(pod_log_file, 'r'))
        for line_dict in the_reader:
            for key, value in line_dict.items():
                if self.__is_float(value):
                    num = float(value)
                    line_dict[key] = round(num, 3)
                elif value == 'n/a':
                    line_dict[key] = 0.0
            line_dict['pod_name'] = pod_name
            workload = self._get_workload_file_name(workload=self._get_run_artifacts_hierarchy(workload_name=workload_name, is_file=True))
            line_dict['run_artifacts_url'] = os.path.join(self._run_artifacts_url, f'{workload}.tar.gz')
            result_list.append(dict(line_dict))
        return result_list

    def _create_vm_run_artifacts(self, vm_name: str, start_stamp: str, end_stamp: str, log_type: str):
        """
        This method creates vm run artifacts
        :param vm_name: vm name
        :param start_stamp: start stamp
        :param end_stamp: end stamp
        :param log_type: log type extension
        :return: run results dict
        """
        result_list = []
        results_list = self._oc.extract_vm_results(vm_name=vm_name, start_stamp=start_stamp, end_stamp=end_stamp)
        workload_name = self._environment_variables_dict.get('workload', '').replace('_', '-')
        # save scale pod log
        if self._scale:
            self._create_pod_log(pod='state-signals-exporter', log_type='.log')
            self._create_pod_log(pod='redis-master', log_type='.log')
        # insert results to csv
        csv_result_file = os.path.join(self._run_artifacts_path, f'{vm_name}{log_type}')
        with open(csv_result_file, 'w') as out:
            for row in results_list:
                if row:
                    out.write(f'{row[0].strip()}\n')
        # csv to dictionary
        the_reader = DictReader(open(csv_result_file, 'r'))
        for line_dict in the_reader:
            for key, value in line_dict.items():
                if self.__is_float(value):
                    num = float(value)
                    line_dict[key] = round(num, 3)
                elif value == 'n/a':
                    line_dict[key] = 0.0
            line_dict['vm_name'] = vm_name
            workload = self._get_workload_file_name(workload=self._get_run_artifacts_hierarchy(workload_name=workload_name, is_file=True))
            line_dict['run_artifacts_url'] = os.path.join(self._run_artifacts_url, f'{workload}.tar.gz')
            result_list.append(dict(line_dict))
        return result_list

    def __make_run_artifacts_tarfile(self, workload: str):
        """
        This method compresses the log file and returns the compressed path
        :return:
        """
        tar_run_artifacts_path = f"{self._run_artifacts_path}.tar.gz"
        with tarfile.open(tar_run_artifacts_path, mode='w:gz') as archive:
            workload_file_name = self._get_workload_file_name(workload)
            archive.add(self._run_artifacts_path, arcname=workload_file_name, recursive=True)
        return tar_run_artifacts_path

    @logger_time_stamp
    def delete_local_artifacts(self):
        """
        This method deletes local artifacts
        :return:
        """
        workload = self._workload.replace('_', '-')
        tar_run_artifacts_path = self.__make_run_artifacts_tarfile(workload)
        # remove local run artifacts workload folder
        # verify that its not empty path
        if len(self._run_artifacts_path) > 3 and self._run_artifacts_path != '/' and self._run_artifacts_path and tar_run_artifacts_path and os.path.isfile(tar_run_artifacts_path):
            # remove run_artifacts_path
            shutil.rmtree(path=self._run_artifacts_path)
            # remove tar.gz file
            os.remove(path=tar_run_artifacts_path)

    @logger_time_stamp
    def upload_run_artifacts_to_s3(self):
        """
        This method uploads log to s3
        :return:
        """
        workload = self._workload.replace('_', '-')
        tar_run_artifacts_path = self.__make_run_artifacts_tarfile(workload)
        run_artifacts_hierarchy = self._get_run_artifacts_hierarchy(workload_name=workload)
        # Upload when endpoint_url is not None
        s3operations = S3Operations()
        # change workload to key convention
        workload_file_name = self._get_workload_file_name(workload)
        upload_file = f"{workload_file_name}.tar.gz"
        s3operations.upload_file(file_name_path=tar_run_artifacts_path,
                                 bucket=self._environment_variables_dict.get('bucket', ''),
                                 key=run_artifacts_hierarchy,
                                 upload_file=upload_file)

    def __get_metadata(self, kind: str = None, status: str = None, result: dict = None) -> dict:
        """
        This method returns metadata for a run, optionally updates by runtime kind
        @param kind: optionally: pod, vm, or kata
        @param status:
        @param result:
        :return:
        """
        date_format = '%Y_%m_%d'
        metadata = {'ocp_version': self._oc.get_ocp_server_version(),
                    'cnv_version': self._oc.get_cnv_version(),
                    'kata_version': self._oc.get_kata_operator_version(),
                    'kata_rpm_version': self._oc.get_kata_rpm_version(node=self._pin_node1),
                    'odf_version': self._oc.get_odf_version(),
                    'runner_version': self._build_version,
                    'version': int(self._build_version.split('.')[-1]),
                    'vm_os_version': 'centos-stream8',
                    'ci_date': datetime.datetime.now().strftime(date_format),
                    'uuid': self._uuid,
                    'pin_node1': self._pin_node1,
                    'pin_node2': self._pin_node2,
                    # display -1 when 0,1 for avoiding conflict with 0/1 status code
                    'odf_disk_count': -1 if self._oc.get_odf_disk_count() in {0, 1} else self._oc.get_odf_disk_count()
}
        if kind:
            metadata.update({'kind': kind})
        if status:
            metadata.update({'run_status': status})
        if self._scale:
            metadata.update({'scale': int(self._scale)*len(self._scale_node_list)})
        if 'bootstorm' in self._workload:
            metadata.update({'vm_os_version': 'fedora37'})
        if 'windows' in self._workload:
            metadata.update({'vm_os_version': self._windows_os})
        if result:
            metadata.update(result)

        return metadata

    def _upload_to_elasticsearch(self, index: str, kind: str, status: str, result: dict = None):
        """
        This method uploads results to elasticsearch
        :param index:
        :param kind:
        :param status:
        :param result:
        :return:
        """
        self._es_operations.upload_to_elasticsearch(index=index, data=self.__get_metadata(kind=kind, status=status, result=result))

    def _verify_elasticsearch_data_uploaded(self, index: str, uuid: str):
        """
        This method verifies that elasticsearch data was uploaded
        :param index:
        :param uuid:
        :return:
        """
        self._es_operations.verify_elasticsearch_data_uploaded(index=index, uuid=uuid)

    @logger_time_stamp
    def update_ci_status(self, status: str, ci_minutes_time: int, benchmark_runner_id: str, benchmark_operator_id: str, benchmark_wrapper_id: str, ocp_install_minutes_time: int = 0, ocp_resource_install_minutes_time: int = 0):
        """
        This method updates ci status Pass/Failed
        :param status: Pass/Failed
        :param ci_minutes_time: ci time in minutes
        :param benchmark_runner_id: benchmark_runner last repository commit id
        :param benchmark_operator_id: benchmark_operator last repository commit id
        :param benchmark_wrapper_id: benchmark_wrapper last repository commit id
        :param ocp_install_minutes_time: ocp install minutes time, default 0 because ocp install run once a week
        :param ocp_resource_install_minutes_time: ocp install minutes time, default 0 because ocp install run once a week
        :return:
        """
        if self._run_type == 'test_ci':
            es_index = 'ci-status-test'
        else:
            es_index = 'ci-status'
        status_dict = {'failed': 0, 'pass': 1}
        metadata = self.__get_metadata()
        if ocp_resource_install_minutes_time != 0:
            ibm_operations = IBMOperations(user=self._environment_variables_dict.get('provision_user', ''))
            ibm_operations.ibm_connect()
            ocp_install_minutes_time = ibm_operations.get_ocp_install_time()
            ibm_operations.ibm_disconnect()
        metadata.update({'status': status, 'status#': status_dict[status], 'ci_minutes_time': ci_minutes_time, 'benchmark_runner_id': benchmark_runner_id, 'benchmark_operator_id': benchmark_operator_id, 'benchmark_wrapper_id': benchmark_wrapper_id, 'ocp_install_minutes_time': ocp_install_minutes_time, 'ocp_resource_install_minutes_time': ocp_resource_install_minutes_time})
        self._es_operations.upload_to_elasticsearch(index=es_index, data=metadata)

    @logger_time_stamp
    def split_run_bulks(self, iterable: range, limit: int = 1):
        """
        This method splits run into bulk depends on threads limit
        @return: run bulks
        """
        length = len(iterable)
        for ndx in range(0, length, limit):
            yield iterable[ndx:min(ndx + limit, length)]

    @logger_time_stamp
    def clear_nodes_cache(self):
        """
        This method clears nodes cache
        """
        for i in range(self.REPEAT_TIMES-1):
            self._oc.clear_node_caches()
            time.sleep(self.SLEEP_TIME)
        self._oc.clear_node_caches()

    def initialize_workload(self):
        """
        This method includes all the initialization of workload
        :return:
        """
        # Verify that CNV operator in installed for CNV workloads
        if '_vm' in self._workload and not self._oc.is_cnv_installed():
            raise CNVNotInstalled(workload=self._workload)
        # Verify that Kata operator in installed for kata workloads
        if '_kata' in self._workload and not self._oc.is_kata_installed():
            raise KataNotInstalled(workload=self._workload)
        if self._delete_all:
            self.delete_all()
            self.clear_nodes_cache()
        if self._odf_pvc:
            self.odf_workload_verification()
        self._template.generate_yamls(scale=str(self._scale), scale_nodes=self._scale_node_list, redis=self._redis, thread_limit=self._threads_limit)
        if self._enable_prometheus_snapshot:
            self.start_prometheus()

    def finalize_workload(self):
        """
        This method includes all the finalization of workload
        :return:
        """
        self._oc.collect_events()
        if self._enable_prometheus_snapshot:
            self.end_prometheus()
        if self._endpoint_url:
            self.upload_run_artifacts_to_s3()
        if not self._save_artifacts_local:
            self.delete_local_artifacts()
        if self._delete_all:
            self.delete_all()




import os
import time
from multiprocessing import Process

from benchmark_runner.common.logger.logger_time_stamp import logger_time_stamp, logger
from benchmark_runner.common.elasticsearch.elasticsearch_exceptions import ElasticSearchDataNotUploaded
from benchmark_runner.workloads.workloads_exceptions import MissingVMs
from benchmark_runner.workloads.workloads_operations import WorkloadsOperations
from benchmark_runner.common.oc.oc import VMStatus


class BootstormVM(WorkloadsOperations):
    """
    This class runs bootstorm vm
    """

    def __init__(self):
        super().__init__()
        self._name = ''
        self._workload_name = ''
        self._es_index = ''
        self._kind = ''
        self._status = ''
        self._vm_name = ''
        self._data_dict = {}
        self._bootstorm_start_time = {}
        # calc total run time - save first vm run time
        self._bootstorm_first_run_time = None

    def save_error_logs(self):
        """
        This method uploads logs into elastic and s3 bucket in case of error
        @return:
        """
        if self._es_host:
            self._data_dict['run_artifacts_url'] = os.path.join(self._run_artifacts_url,
                                                                f'{self._get_run_artifacts_hierarchy(workload_name=self._get_workload_file_name(self._workload_name), is_file=True)}.tar.gz')
            self._upload_to_elasticsearch(index=self._es_index, kind=self._kind, status='failed',
                                          result=self._data_dict)
            # verify that data upload to elastic search according to unique uuid
            self._verify_elasticsearch_data_uploaded(index=self._es_index, uuid=self._uuid)

    @logger_time_stamp
    def _set_bootstorm_vm_first_run_time(self):
        """
        This method sets the first vm run time
        @return:
        """
        self._bootstorm_first_run_time = time.time()

    @logger_time_stamp
    def _get_bootstorm_vm_total_run_time(self):
        """
        This method retrieves the total run time from the first VM execution
        @return: The delta time from the first VM execution
        """
        delta = time.time() - self._bootstorm_first_run_time
        return round(delta, 3) * self.MILLISECONDS

    @logger_time_stamp
    def _set_bootstorm_vm_start_time(self, vm_name: str = ''):
        """
        This method captures boot start time for specified VM
        @return:
        """
        self._bootstorm_start_time[vm_name] = time.time()

    @logger_time_stamp
    def _ssh_vm(self, vm_name: str):
            """
            Verify ssh into VM and return vm node in success or False if failed
            @return:
            """
            self._virtctl.expose_vm(vm_name=vm_name)
            # wait till vm ssh login
            if self._oc.get_vm_node(vm_name=vm_name):
                vm_node = self._oc.get_vm_node(vm_name=vm_name)
                if vm_node:
                    node_ip = self._oc.get_nodes_addresses()[vm_node]
                    vm_node_port = self._oc.get_exposed_vm_port(vm_name=vm_name)
                    if self._oc.wait_for_vm_ssh(vm_name=vm_name, node_ip=node_ip, vm_node_port=vm_node_port):
                        logger.info(f"Successfully ssh into VM: '{vm_name}' in Node: '{vm_node}' ")
                    return vm_node
            return False

    @logger_time_stamp
    def _get_bootstorm_vm_elapsed_time(self, vm_name: str, vm_node: str) -> dict:
        """
        Returns the boot elapse time for the specified VM in milliseconds.
        @return: Dictionary with vm_name, node, and boot elapse time.
        """
        if vm_node:
            delta = round((time.time() - self._bootstorm_start_time[vm_name]) * self.MILLISECONDS, 3)
            return {'vm_name': vm_name, 'node': vm_node, 'bootstorm_time': delta, 'vm_ssh': int(bool(vm_node)),}
        return {}

    def _create_vm_scale(self, vm_num: str):
        """
        This method creates VMs in parallel
        """
        try:
            self._oc.create_async(yaml=os.path.join(f'{self._run_artifacts_path}', f'{self._name}_{vm_num}.yaml'))
            self._oc.wait_for_vm_status(vm_name=f'{self._workload_name}-{self._trunc_uuid}-{vm_num}', status=VMStatus.Stopped)
        except Exception as err:
            # save run artifacts logs
            self.save_error_logs()
            raise err

    def _finalize_vm(self):
        self._status = 'complete' if self._data_dict else 'failed'
        # prometheus queries
        self._prometheus_metrics_operation.finalize_prometheus()
        metric_results = self._prometheus_metrics_operation.run_prometheus_queries()
        prometheus_result = self._prometheus_metrics_operation.parse_prometheus_metrics(data=metric_results)
        # update total vm run time
        if not self._verification_only:
            total_run_time = self._get_bootstorm_vm_total_run_time()
            self._data_dict.update({'total_run_time': total_run_time})
        self._data_dict.update(prometheus_result)
        if self._es_host:
            # upload several run results
            self._upload_to_elasticsearch(index=self._es_index, kind=self._kind, status=self._status,
                                          result=self._data_dict)
            # verify that data upload to elastic search according to unique uuid
            self._verify_elasticsearch_data_uploaded(index=self._es_index, uuid=self._uuid)

    def _run_vm(self):
        """
        This method runs one VM, upload results to Elasticsearch, and destroys VM synchronously
        @return:
        """
        self._oc.create_async(yaml=os.path.join(f'{self._run_artifacts_path}', f'{self._name}.yaml'))
        self._oc.wait_for_vm_status(vm_name=f'{self._workload_name}-{self._trunc_uuid}', status=VMStatus.Stopped)
        self._set_bootstorm_vm_first_run_time()
        self._set_bootstorm_vm_start_time(vm_name=self._vm_name)
        self._virtctl.start_vm_sync(vm_name=self._vm_name)
        self.vm_node = self._ssh_vm(vm_name=self._vm_name)
        self._data_dict = self._get_bootstorm_vm_elapsed_time(vm_name=self._vm_name, vm_node=self.vm_node)
        self._data_dict['run_artifacts_url'] = os.path.join(self._run_artifacts_url,
                                                            f'{self._get_run_artifacts_hierarchy(workload_name=self._workload_name, is_file=True)}-{self._time_stamp_format}.tar.gz')
        self._finalize_vm()
        self._oc.delete_vm_sync(
            yaml=os.path.join(f'{self._run_artifacts_path}', f'{self._name}.yaml'),
            vm_name=self._vm_name)

    def _verify_vm_ssh(self):
        """
        This method verifies each VM ssh login
        :return:
        """
        try:
            vm_names = self._oc._get_all_vm_names()
            if not vm_names:
                raise MissingVMs
            for vm_name in vm_names:
                vm_node = self._ssh_vm(vm_name)
                self._data_dict = {
                    'vm_name': vm_name,
                    'node': vm_node,
                    'vm_ssh': int(bool(vm_node)),
                    'run_artifacts_url': os.path.join(
                        self._run_artifacts_url,
                        f"{self._get_run_artifacts_hierarchy(workload_name=self._workload_name, is_file=True)}-{self._time_stamp_format}.tar.gz"
                    )
                }
                self._finalize_vm()
        except Exception as err:
            # save run artifacts logs
            self.save_error_logs()
            raise err

    def _run_vm_scale(self, vm_num: str):
        """
        This method runs VMs in parallel and wait for login to be enabled
        """
        try:
            vm_name = f'{self._workload_name}-{self._trunc_uuid}-{vm_num}'
            self._set_bootstorm_vm_start_time(vm_name=f'{self._workload_name}-{self._trunc_uuid}-{vm_num}')
            self._virtctl.start_vm_async(vm_name=f'{self._workload_name}-{self._trunc_uuid}-{vm_num}')
            self._virtctl.wait_for_vm_status(vm_name=vm_name, status=VMStatus.Running)
            vm_node = self._ssh_vm(vm_name)
            self._data_dict = self._get_bootstorm_vm_elapsed_time(vm_name=vm_name, vm_node=vm_node)
            self._data_dict['run_artifacts_url'] = os.path.join(self._run_artifacts_url, f'{self._get_run_artifacts_hierarchy(workload_name=self._workload_name, is_file=True)}-scale-{self._time_stamp_format}.tar.gz')
            self._finalize_vm()
        except Exception as err:
            # save run artifacts logs
            self.save_error_logs()
            raise err

    def _stop_vm_scale(self, vm_num: str):
        """
        This method stops VMs async in parallel
        """
        try:
            self._virtctl.stop_vm_async(vm_name=f'{self._workload_name}-{self._trunc_uuid}-{vm_num}')
        except Exception as err:
            # save run artifacts logs
            self.save_error_logs()
            raise err

    def _wait_for_stop_vm_scale(self, vm_num: str):
        """
        This method waits for VMs stop in parallel
        """
        try:
            self._virtctl.wait_for_vm_status(vm_name=f'{self._workload_name}-{self._trunc_uuid}-{vm_num}')
        except Exception as err:
            # save run artifacts logs
            self.save_error_logs()
            raise err

    def _delete_vm_scale(self, vm_num: str):
        """
        This method deletes VMs async in parallel
        """
        try:
            self._oc.delete_async(
                yaml=os.path.join(f'{self._run_artifacts_path}', f'{self._name}_{vm_num}.yaml'))
        except Exception as err:
            # save run artifacts logs
            self.save_error_logs()
            raise err

    def _wait_for_delete_vm_scale(self, vm_num: str):
        """
        This method waits for VMs delete in parallel
        """
        try:
            self._oc.wait_for_vm_delete(vm_name=f'{self._workload_name}-{self._trunc_uuid}-{vm_num}')
        except Exception as err:
            # save run artifacts logs
            self.save_error_logs()
            raise err

    def _initialize_run(self):
        """
        Initialize prometheus start time, vm name, kind and create benchmark-runner namespace for bootstorm vms
        """
        self._prometheus_metrics_operation.init_prometheus()
        self._name = self._workload
        self._workload_name = self._workload.replace('_', '-')
        self._vm_name = f'{self._workload_name}-{self._trunc_uuid}'
        self._kind = 'vm'
        self._environment_variables_dict['kind'] = 'vm'
        if not self._verification_only:
            # create namespace
            self._oc.create_async(yaml=os.path.join(f'{self._run_artifacts_path}', 'namespace.yaml'))

    def run_vm_workload(self):
        # verification only w/o running or deleting any resource
        if self._verification_only:
            self._verify_vm_ssh()
        else:
            if not self._scale:
                self._run_vm()
            # scale
            else:
                first_run_time_updated = False
                # Run VMs only
                if not self._delete_all:
                    steps = (self._create_vm_scale, self._run_vm_scale)
                else:
                    steps = (self._create_vm_scale, self._run_vm_scale, self._stop_vm_scale,
                             self._wait_for_stop_vm_scale,self._delete_vm_scale, self._wait_for_delete_vm_scale)

                # create run bulks
                bulks = tuple(self.split_run_bulks(iterable=range(self._scale * len(self._scale_node_list)),
                                                   limit=self._threads_limit))
                # create, run and delete vms
                for target in steps:
                    proc = []
                    for bulk in bulks:
                        for vm_num in bulk:
                            # save the first run vm time
                            if self._run_vm_scale == target and not first_run_time_updated:
                                self._set_bootstorm_vm_first_run_time()
                                first_run_time_updated = True
                            p = Process(target=target, args=(str(vm_num),))
                            p.start()
                            proc.append(p)
                        for p in proc:
                            p.join()
                        # sleep between bulks
                        time.sleep(self._bulk_sleep_time)
                        proc = []

    @logger_time_stamp
    def run(self):
        """
        This method runs the workload
        :return:
        """
        try:
            self._initialize_run()
            if self._run_type == 'test_ci':
                self._es_index = 'bootstorm-test-ci-results'
            else:
                self._es_index = 'bootstorm-results'
            self.run_vm_workload()
            # delete namespace
            self._oc.delete_async(yaml=os.path.join(f'{self._run_artifacts_path}', 'namespace.yaml'))
        except ElasticSearchDataNotUploaded as err:
            self._oc.delete_vm_sync(
                yaml=os.path.join(f'{self._run_artifacts_path}', f'{self._name}.yaml'),
                vm_name=self._vm_name)
            raise err
        except Exception as err:
            # save run artifacts logs
            self.save_error_logs()
            raise err




from benchmark_runner.common.grafana.grafana_operations import GrafanaOperations
from benchmark_runner.main.environment_variables import environment_variables

# Update grafana perf dashboard
environment_variables_dict = environment_variables.environment_variables_dict
grafana_operations = GrafanaOperations(grafana_url=environment_variables_dict.get('grafana_url', ''),
                                       grafana_api_key=environment_variables_dict.get('grafana_api_key', ''),
                                       grafana_json_path=environment_variables_dict.get('grafana_json_path', ''),
                                       grafana_folder_name=environment_variables_dict.get('grafana_folder_name', ''))
# for debug: grafana_operations.fetch_all_dashboards()
grafana_operations.read_dashboard_json()
grafana_operations.increment_dashboard_version()
grafana_operations.override_dashboard()




import pandas as pd


class AnalyzeData:
    """
    This class analyzing input data using pandas
    """
    def __init__(self, input_data_path):
        self.__input_data_path = input_data_path
        self._runs_list = []

    def analyze_vdbench_csv_data(self):
        """
        This method read input vdbench CSV data into pandas dataframe
        :return: list of dataframes, first one is the average results
        """
        run_dict = {}
        run_metrics_dict = {}
        df = pd.read_csv(self.__input_data_path)
        df['timestamp'] = pd.to_datetime(df['timestamp'], format='%m/%d/%Y-%H:%M:%S-PDT')
        df.rename(columns={"timestamp": "time"})
        df_groupbyrun = df.groupby(df.columns[2]).mean()
        # df filter out all run that start with the letter 'f'
        df_run_avg = df_groupbyrun[~df_groupbyrun.index.str.startswith('f')]
        self._runs_list = df_run_avg.index.to_list()
        df_run_avg = df_run_avg.round(2)
        df_run_avg = df_run_avg.fillna(0)
        avg_run_dict = df_run_avg.to_dict(orient='index')

        # create run df dict by selected run name
        for run_name in self._runs_list:
            df_run = df[df.iloc[:, 2] == run_name].round(2)
            df_run = df_run.fillna(0)
            run_dict[run_name] = df_run.to_dict()

        # filter highlight metrics
        for run in self._runs_list:
            run_metrics_dict[run + '_cpu_used'] = run_dict[run]['cpu_used']

        return avg_run_dict, run_metrics_dict

    @property
    def runs_list(self):
        """
        This method return runs list
        :return:
        """
        return self._runs_list




import os
from typeguard import typechecked

# display bokeh
from bokeh.plotting import figure, show
from bokeh.io import output_notebook, output_file
from bokeh.models import HoverTool, Label, ColumnDataSource, LabelSet, Circle, NumeralTickFormatter


from benchmark_runner.jupyterlab.templates.logs_operations.logs_operations import LogsOperations
from benchmark_runner.jupyterlab.templates.analyze_workloads.visualize_workload_operations import VisualizeWorkloadOperations


class AnalyzeHammerdbLogs(VisualizeWorkloadOperations):
    """
    This class analyzes Hammerdb logs
    """

    def __init__(self, s3_logs_url: str):
        super().__init__()
        self.__workload = 'hammerdb'
        self.s3_logs_url = s3_logs_url
        self.__logs_operations = LogsOperations(s3_logs_url=self.s3_logs_url)

    def get_hammerdb_log_file(self):
        """
        This method returns hammerdb log file
        @return: hammerdb log file
        """
        workload_dir = self.__logs_operations.get_workload_dir()
        workload_files = [file for file in os.listdir(workload_dir) if file.startswith(self.__workload)]
        # Iterate over all files in the directory
        for file_name in workload_files:
            if 'workload' in file_name:
                workload_log_file = os.path.join(workload_dir, file_name)
        return workload_log_file

    @typechecked
    def extract_hammerdb_result(self, workload_log_file: str):
        """
        This method extracts hammerdb result
        @return: hammerdb result
        """
        start_content = '+-------------------------------------------------- HammerDB Results --------------------------------------------------+'
        end_content = '+-------------------------------------------------------------------------------------------------------------------+'
        with open(workload_log_file, 'r') as file:
            hammerdb_result = {}
            should_print = False
            for line in file:
                line = line.strip()
                if start_content in line:
                    should_print = True
                elif end_content in line:
                    should_print = False
                elif should_print:
                    if 'Worker(s):' in line:
                        thread = line.split(':')[-1]
                    if 'TPM:' in line:
                        tpm = line.split(':')[-1]
                        hammerdb_result[thread] = tpm
        return hammerdb_result

    @typechecked
    def display_hammerdb_result(self, hammerdb_result: dict, workload_log_file: str):
        """
        This method displays hammerdb result
        @return:
        """
        # Convert dictionary keys and values to lists
        x = list(hammerdb_result.keys())
        y = [int(val) for val in hammerdb_result.values()]

        # Create a figure
        p = figure(x_range=x, title=workload_log_file.split('/')[-1], x_axis_label='Threads', y_axis_label='TPM')

        # Set y-axis type to display full numbers
        p.yaxis.axis_label_text_font_size = '10pt'
        p.yaxis.axis_label_text_font_style = 'normal'
        p.yaxis.formatter = NumeralTickFormatter(format='0,0')
        p.yaxis.axis_label_standoff = 10

        # Add vertical bars to the figure with wider width
        bars = p.vbar(x=x, top=y, width=0.8)

        # Calculate the lower position for the labels
        label_offset = max(y) * 0.01
        label_y = [val - label_offset for val in y]

        # Add labels below the bars with commas
        p.text(x=x, y=label_y, text=[f"{val:,}" for val in y], text_baseline="top", text_align="center",
               text_font_size='10pt', text_color='white')

        # Add hover tool with x-axis and y-axis values
        hover = HoverTool(tooltips=[('Threads', '@x'), ('TPM', '@top{0,0}')], renderers=[bars])
        p.add_tools(hover)

        # Specify the output file
        output_file(os.path.join(f"{workload_log_file.split('.')[0]}.html"))

        # Display the plot in the notebook
        output_notebook()
        show(p)

    def compare_results(self, result1: dict, result2: dict, legend_label1: str, legend_label2: str, database: str):
        """
        This method compare between 2 hammerdb results
        @param result1:
        @param result2:
        @param legend_label1:
        @param legend_label2:
        @param database: PostgreSQL, Mariadb, MSSQL
        @return:
        """
        self.compare_run_results(result1=result1, result2=result2, legend_label1=legend_label1, legend_label2=legend_label2, title=f'{database} KTPM [ KTransactions Per Minutes ]', x_axis_label = 'Threads', y_axis_label= 'KTPM')




import os
import time
import yaml
import sys
import benchmark_runner
from multiprocessing import Process

from jinja2 import Template, TemplateSyntaxError
from benchmark_runner.common.template_operations.render_yaml_from_template import render_yaml_file
from benchmark_runner.common.logger.logger_time_stamp import logger_time_stamp, logger
from benchmark_runner.main.environment_variables import environment_variables


class TemplateOperations:
    """This class is responsible for template operations"""

    def __init__(self, workload: str = ''):
        self.__workload = workload
        self.initialize_environment(environment_variables.environment_variables_dict)

    def __initialize_dependent_variables__(self):
        self.__run_type = self.__environment_variables_dict.get('run_type', '')
        self.__run_artifacts_path = self.__environment_variables_dict.get('run_artifacts_path', '')
        self.__bulk_sleep_time = int(self.__environment_variables_dict.get('bulk_sleep_time', ''))
        self.__dir_path = os.path.abspath(os.path.join(os.path.dirname(__file__), 'templates'))
        if self.__run_type == 'test_ci':
            self.__environment_variables_dict['es_suffix'] = '-test-ci'
        else:
            self.__environment_variables_dict['es_suffix'] = ''

    @logger_time_stamp
    def __split_run_bulks(self, iterable: range, limit: int = 1):
        """
        This method splits run into bulk depends on threads limit
        @return: run bulks
        """
        length = len(iterable)
        for ndx in range(0, length, limit):
            yield iterable[ndx:min(ndx + limit, length)]

    def __get_workload_template_kind(self):
        """
        Kata shares templates with pod
        """
        if self.__workload_kind == 'kata' or self.__workload_kind == 'pod':
            if self.__environment_variables_dict.get('kata_cpuoffline_workaround', False):
                logger.warn('*** WARNING: Enabling Kata CPU offline workaround ***')
            return 'pod'
        elif self.__workload_kind == 'vm':
            return 'vm'
        else:
            return None

    @staticmethod
    def __get_sub_dict(dictionary: dict, key: str, value: str):
        """
        Return a sub-dictionary of input dictionary as specified below.
        If dictionary[key][value] exists, return that.
        If dictionary[key][value] does not exist, but dictionary[key]['default'] exists, return
           dictionary[key]['default']
        Otherwise return an empty dictionary.
        :param dictionary: dictionary to be searched
        :param key: key to search for in dictionary
        :param value: value to be searched for under key
        :return: dictionary[key][value] if that exists, dictionary[key]['default'] if that exists, otherwise empty dictionary.
        """
        if key in dictionary:
            subdict = dictionary[key]
            if value in subdict:
                return subdict[value]
            elif 'default' in subdict:
                return subdict['default']
        return dict()

    def __build_template_data(self, previous_data: dict, template_root_data: dict):
        """
        Build data dictionary for rendering from template data.
        """
        template_data = template_root_data.get('template_data')
        shared_data = template_data.get('shared')
        if shared_data is None:
            shared_data = {}
        kind_data = self.__get_sub_dict(template_data, 'kind', self.__workload_kind)
        run_type_data = self.__get_sub_dict(template_data, 'run_type', self.__run_type)
        kind_runtype_data = self.__get_sub_dict(kind_data, 'run_type', self.__run_type)
        extra_data = self.__get_sub_dict(template_data, 'extra', self.__workload_extra_name)
        extra_kind_data = self.__get_sub_dict(extra_data, 'kind', self.__workload_kind)
        extra_runtype_data = self.__get_sub_dict(extra_data, 'run_type', self.__run_type)
        extra_kind_runtype_data = self.__get_sub_dict(extra_kind_data, 'run_type', self.__run_type)

        return {**previous_data, **shared_data,
                **kind_data, **run_type_data, **kind_runtype_data,
                **extra_data, **extra_kind_data, **extra_runtype_data, **extra_kind_runtype_data}

    @logger_time_stamp
    def __generate_yamls_internal(self, scale: str = None, scale_num: str = None, scale_node: str = None, redis: str = None):
        """
        Generate required .yaml files as a dictionary of file names and contents
        @param scale: current scale 0,1,2
        @param scale_num: the scale number
        @param scale_node: the scale node
        @param redis: redis
        """
        # This really belongs in __init__, but some tests rely on
        # being able to create this without an actual workload
        answer = {}
        self.__workload_name = self.__workload.split('_')[0]
        self.__workload_kind = self.__workload.split('_')[1]
        self.__workload_extra_name = '_'.join(self.__workload.split('_')[2:3])
        self.__workload_template_kind = self.__get_workload_template_kind()
        if self.__workload_extra_name:
            self.__standard_output_file = f"{'_'.join([self.__workload_name, self.__workload_template_kind, self.__workload_extra_name])}.yaml"
        else:
            if scale:
                self.__standard_output_file = f"{'_'.join([self.__workload_name, self.__workload_template_kind, scale])}.yaml"
            else:
                self.__standard_output_file = f"{'_'.join([self.__workload_name, self.__workload_template_kind])}.yaml"
        self.__standard_template_file = f"{'_'.join([self.__workload_name, self.__workload_template_kind])}_template.yaml"
        workload_dir_path = os.path.join((os.path.dirname(benchmark_runner.__file__)), "workloads", self.__workload, "template") \
            if self.__environment_variables_dict['template_in_workload_dir'] \
               else os.path.join(self.__dir_path, self.__workload_name)
        template_render_data = {
            'kind': self.__workload_kind,
            'workload_name': self.__workload_name,
            'workload_template_kind': self.__workload_template_kind,
            'workload_extra_name': self.__workload_extra_name,
            'standard_output_file': self.__standard_output_file,
            'standard_template_file': self.__standard_template_file,
            'scale': scale,
            'scale_num': scale_num,
            'scale_node': scale_node
            }
        self.__environment_variables_dict = {**self.__environment_variables_dict, **template_render_data}
        common_data = yaml.load(render_yaml_file(dir_path=self.__dir_path, yaml_file='common_template.yaml', environment_variable_dict=self.__environment_variables_dict), Loader=yaml.FullLoader)['common_data']

        template_render_data = {**self.__environment_variables_dict, **common_data}

        workload_data = yaml.load(render_yaml_file(workload_dir_path, f'{self.__workload_name}_data_template.yaml', template_render_data), Loader=yaml.FullLoader)

        if self.__environment_variables_dict.get("config_from_args"):
            config = dict([kv.split("=") for kv in sys.argv[1:]])
            workload_data["template_data"]["run_type"]["default"] = config

        render_data = self.__build_template_data(template_render_data, workload_data)

        out_files = [{'name': self.__standard_output_file, 'template': self.__standard_template_file}]
        if 'files' in workload_data:
            out_files.extend(workload_data['files'])
        for out_file in out_files:
            filename = out_file['name']
            if 'template' in out_file:
                template_file = out_file['template']
            else:
                file_components = os.path.splitext(filename)
                template_file = f'{file_components[0]}_template{file_components[1]}'
            try:
                with open(os.path.join(workload_dir_path, 'internal_data', template_file)) as f:
                    template = Template(f.read(), keep_trailing_newline=True)
            except TemplateSyntaxError as err:
                raise SyntaxError(f"Error while rendering {template_file}: {err}")
            answer[filename] = f"{template.render(render_data)}"
            # namespace
            if os.path.isfile(os.path.join(self.__dir_path, f'namespace_template.yaml')):
                answer['namespace.yaml'] = render_yaml_file(dir_path=self.__dir_path, yaml_file='namespace_template.yaml', environment_variable_dict=self.__environment_variables_dict)
            # windows workload
            if 'windows' in self.__workload_name:
                answer['windows_dv.yaml'] = render_yaml_file(dir_path=os.path.join(workload_dir_path, 'internal_data'), yaml_file='windows_dv_template.yaml', environment_variable_dict=render_data)
            # vdbench scale
            if scale and redis and 'vdbench' in self.__workload_name:
                    answer['redis.yaml'] = render_yaml_file(dir_path=os.path.join(self.__dir_path, 'scale'), yaml_file='redis_template.yaml', environment_variable_dict=self.__environment_variables_dict)
                    answer['state_signals_exporter_pod.yaml'] = render_yaml_file(dir_path=os.path.join(self.__dir_path, 'scale'), yaml_file='state_signals_exporter_pod_template.yaml', environment_variable_dict=self.__environment_variables_dict)
        for filename, data in answer.items():
            with open(os.path.join(self.__run_artifacts_path, filename), 'w') as f:
                f.write(data)

    @logger_time_stamp
    def generate_yamls(self, scale: str = '', scale_nodes: list = [], redis: str = None, thread_limit: int = None):
        """
        This method generate workload yaml from template
        :return:
        """
        if not scale:
            self.__generate_yamls_internal()
        else:
            proc = []
            scale = int(scale)
            bulks = tuple(self.__split_run_bulks(iterable=range(scale * len(scale_nodes)), limit=thread_limit))
            if scale_nodes:
                for bulk in bulks:
                    for num in bulk:
                        p = Process(target=self.__generate_yamls_internal, args=(str(num), str(scale * len(scale_nodes)), scale_nodes[int(num / scale)], redis,))
                        p.start()
                        proc.append(p)
                    for p in proc:
                        p.join()
                    # sleep between bulks
                    time.sleep(self.__bulk_sleep_time)
            else:
                for scale_num in range(scale):
                    self.__generate_yamls_internal(scale=str(scale_num))

    # The following routines are for testing purposes,
    # in particular to allow a known environment to be set up for golden file testing
    def initialize_environment(self, environment: dict = environment_variables.environment_variables_dict):
        self.__environment_variables_dict = environment
        self.__initialize_dependent_variables__()

    def set_environment_variable(self, variable: str, value: str):
        self.__environment_variables_dict[variable] = value
        self.__initialize_dependent_variables__()

    def set_environment_variables(self, update_environment: dict):
        self.__environment_variables_dict = {**self.__environment_variables_dict, **update_environment}
        self.__initialize_dependent_variables__()

    def clear_environment_variable(self, variable):
        self.__environment_variables_dict.pop(variable, '')
        self.__initialize_dependent_variables__()

    def get_current_run_path(self):
        return self.__run_artifacts_path




import subprocess
from benchmark_runner.common.logger.logger_time_stamp import logger
from benchmark_runner.common.ssh.ssh_exceptions import SSHSubprocessError
from benchmark_runner.main.environment_variables import environment_variables


class SSH:
    """
    This class run local SSH commands
    """
    def __init__(self):
        self.__environment_variables_dict = environment_variables.environment_variables_dict
        self.timeout = int(self.__environment_variables_dict.get('timeout', 300))

    def run(self, cmd: str, is_check: bool = False, background: bool = False):
        """
        This method run shell commands
        :param background: run ssh in background
        :param cmd:
        :param is_check:run check command
        :return:
        """
        try:
            if background:
                output = subprocess.Popen(cmd, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
            elif is_check:
                output = subprocess.check_output(
                    cmd,
                    stderr=subprocess.STDOUT,
                    shell=True, timeout=self.timeout,
                    universal_newlines=False)
            # execute cmd
            else:
                output = subprocess.getoutput(cmd)
            return output
        except subprocess.CalledProcessError as err:
            logger.error(f'subprocess Status : FAIL: {err.returncode} {err.output}')
            raise SSHSubprocessError()
        except Exception as err:
             raise err




import requests
import json
import logging

logging.basicConfig(level=logging.INFO)


class GrafanaOperations:
    """
    This class is responsible for Grafana operations
    """
    def __init__(self, grafana_url: str, grafana_api_key: str, grafana_json_path: str, grafana_folder_name: str):
        self.grafana_url = grafana_url
        self.grafana_api_key = grafana_api_key
        self.grafana_json_path = grafana_json_path
        self.dashboard_data = {}
        self.logger = logging.getLogger(__name__)
        self.__grafana_folder_name = grafana_folder_name

    def fetch_all_dashboards(self):
        """
        This method fetches all dashboards
        :return:
        """
        dashboard_list = []
        headers = {
            "Authorization": f"Bearer {self.grafana_api_key}",
        }

        try:
            response = requests.get(f"{self.grafana_url}/api/search", headers=headers)

            if response.status_code == 200:
                dashboards = response.json()
                for dashboard in dashboards:
                    dashboard_list.append(f"Dashboard ID: {dashboard['id']}, Title: {dashboard['title']}")
            else:
                raise Exception(f"Failed to fetch dashboards. Status code: {response.status_code}. Message: {response.text}")
            return dashboard_list
        except requests.exceptions.RequestException as e:
            raise Exception(f"Error fetching dashboards: {e}")

    def increment_dashboard_version(self):
        """
        This method increments dashboard version
        :return:
        """
        self.dashboard_data["version"] = self.get_latest_dashboard_version()

    def read_dashboard_json(self):
        """
        This method reads dashboard from json into dictionary
        :return:
        """
        with open(self.grafana_json_path, 'r') as f:
            self.dashboard_data = json.load(f)

    def write_dashboard_json(self):
        """
        This method writes dashboard data into json
        :return:
        """
        with open(self.grafana_json_path, 'w') as json_file:
            json.dump(self.dashboard_data, json_file, indent=2)

    def get_latest_dashboard_version(self):
        """
        This method get latest dashboard version
        :return:
        """
        headers = {
            "Authorization": f"Bearer {self.grafana_api_key}",
        }
        try:
            response = requests.get(f"{self.grafana_url}/api/dashboards/uid/{self.dashboard_data['uid']}", headers=headers)
            response_json = response.json()
            return response_json['dashboard']['version']

        except requests.exceptions.RequestException as e:
            raise Exception(f"Error fetching dashboard version: {e}")

    def get_dashboard_folder_id(self):
        """
        This method gets dashboard folder id according to grafana folder name
        """
        headers = {
            "Authorization": f"Bearer {self.grafana_api_key}",
            "Content-Type": "application/json",
        }

        try:
            response = requests.get(
                f"{self.grafana_url}/api/search",
                headers=headers
            )

            if response.status_code == 200:
                data = response.json()
                for item in data:
                    if item.get('type') == 'dash-folder' and item.get('title') == self.__grafana_folder_name:
                        return item['id']

            raise Exception(f"Folder with name '{self.__grafana_folder_name}' not found.")

        except requests.exceptions.RequestException as e:
            raise Exception(f"Error retrieving folder ID: {e}")

    def override_dashboard(self):
        """
        This method overrides dashboard with new json
        :return:
        """
        headers = {
            "Authorization": f"Bearer {self.grafana_api_key}",
            "Content-Type": "application/json",
        }
        try:
            response = requests.post(
                f"{self.grafana_url}/api/dashboards/db",
                headers=headers,
                json={
                    "dashboard": self.dashboard_data,
                    "folderId": self.get_dashboard_folder_id()  # Specify the folder ID here
                },
            )

            if response.status_code == 200:
                self.logger.info(f"Dashboard '{self.dashboard_data['title']}' overridden successfully.")
            else:
                # The 412 status code is used when a newer dashboard already exists.
                raise Exception(
                    f"Failed to override dashboard '{self.dashboard_data['title']}'. Status code: {response.status_code}. Message: {response.text}")

        except requests.exceptions.RequestException as e:
            raise Exception(f"Error overriding dashboard '{self.dashboard_data['title']}': {e}")
